
from __future__ import annotations

import asyncio
import os.path
import uuid
from argparse import ArgumentParser, Namespace
from contextlib import closing, asynccontextmanager
from datetime import datetime, timezone
from secrets import token_bytes
from typing import Any, Optional, Tuple, Sequence, Mapping, Dict, Awaitable, \
    AsyncIterator
from typing_extensions import Final

from pkg_resources import resource_listdir, resource_stream
from pysasl.creds import AuthenticationCredentials
from pysasl.hashing import Cleartext

from pymap.config import BackendCapability, IMAPConfig
from pymap.exceptions import AuthorizationFailure, NotAllowedError, \
    UserNotFound
from pymap.interfaces.backend import BackendInterface, ServiceInterface
from pymap.interfaces.login import LoginTokenData, LoginInterface, \
    IdentityInterface
from pymap.parsing.message import AppendMessage
from pymap.parsing.specials.flag import Flag, Recent
from pymap.user import UserMetadata

from .filter import FilterSet
from .mailbox import Message, MailboxData, MailboxSet
from ..session import BaseSession

__all__ = ['DictBackend', 'Config']


class DictBackend(BackendInterface):
    """Defines a backend that uses an in-memory dictionary for example usage
    and integration testing.

    """

    def __init__(self, login: Login, config: Config) -> None:
        super().__init__()
        self._login = login
        self._config = config

    @property
    def login(self) -> Login:
        return self._login

    @property
    def config(self) -> Config:
        return self._config

    @classmethod
    def add_subparser(cls, name: str, subparsers: Any) -> ArgumentParser:
        parser = subparsers.add_parser(name, help='in-memory backend')
        parser.add_argument('--demo-data', action='store_true',
                            help='load initial demo data')
        parser.add_argument('--demo-user', default='demouser',
                            metavar='VAL', help='demo user ID')
        parser.add_argument('--demo-password', default='demopass',
                            metavar='VAL', help='demo user password')
        return parser

    @classmethod
    async def init(cls, args: Namespace) -> Tuple[DictBackend, Config]:
        config = Config.from_args(args)
        login = Login(config)
        return cls(login, config), config

    async def start(self, services: Sequence[ServiceInterface]) -> Awaitable:
        tasks = [await service.start() for service in services]
        return asyncio.gather(*tasks)


class Config(IMAPConfig):
    """The config implementation for the dict backend."""

    def __init__(self, args: Namespace, *, demo_data: bool,
                 demo_user: str, demo_password: str,
                 demo_data_resource: str = __name__, **extra: Any) -> None:
        super().__init__(args, hash_context=Cleartext(), **extra)
        self._demo_data = demo_data
        self._demo_user = demo_user
        self._demo_password = demo_password
        self._demo_data_resource = demo_data_resource
        self.set_cache: Dict[str, Tuple[MailboxSet, FilterSet]] = {}

    @property
    def backend_capability(self) -> BackendCapability:
        return BackendCapability(idle=True, object_id=True, multi_append=True)

    @property
    def demo_data(self) -> bool:
        """True if demo data should be loaded at startup."""
        return self._demo_data

    @property
    def demo_data_resource(self) -> str:
        """Resource path of demo data files."""
        return self._demo_data_resource

    @property
    def demo_user(self) -> str:
        """A login name that is valid at startup, which defaults to
        ``demouser``.

        """
        return self._demo_user

    @property
    def demo_password(self) -> str:
        """The password for the :attr:`.demo_user` login name, which defaults
        to ``demopass``.

        """
        return self._demo_password

    @classmethod
    def parse_args(cls, args: Namespace) -> Mapping[str, Any]:
        return {**super().parse_args(args),
                'demo_data': args.demo_data,
                'demo_user': args.demo_user,
                'demo_password': args.demo_password}


class Session(BaseSession[Message]):
    """The session implementation for the dict backend."""

    def __init__(self, owner: str, config: Config, mailbox_set: MailboxSet,
                 filter_set: FilterSet) -> None:
        super().__init__(owner)
        self._config = config
        self._mailbox_set = mailbox_set
        self._filter_set = filter_set

    @property
    def config(self) -> Config:
        return self._config

    @property
    def mailbox_set(self) -> MailboxSet:
        return self._mailbox_set

    @property
    def filter_set(self) -> FilterSet:
        return self._filter_set


class Login(LoginInterface):
    """The login implementation for the dict backend."""

    def __init__(self, config: Config) -> None:
        super().__init__()
        self.config = config
        self.users = {config.demo_user: UserMetadata(
            config, password=config.demo_password)}
        self.tokens: Dict[str, Tuple[str, bytes]] = {}

    async def authenticate(self, credentials: AuthenticationCredentials) \
            -> Identity:
        authcid = credentials.authcid
        token_key: Optional[bytes] = None
        role: Optional[str] = None
        if credentials.authcid_type == 'login-token':
            if authcid in self.tokens:
                authcid, token_key = self.tokens[authcid]
        elif credentials.authcid_type == 'admin-token':
            authcid = credentials.identity
            role = 'admin'
        try:
            metadata = await Identity(authcid, self, None).get()
        except UserNotFound:
            metadata = UserMetadata(self.config)
        await metadata.check_password(credentials, token_key=token_key)
        role = role or metadata.role
        if role != 'admin' and authcid != credentials.identity:
            raise AuthorizationFailure()
        return Identity(credentials.identity, self, role)


class Identity(IdentityInterface):
    """The identity implementation for the dict backend."""

    def __init__(self, name: str, login: Login, role: Optional[str]) -> None:
        super().__init__()
        self.login: Final = login
        self.config: Final = login.config
        self._name = name
        self._role = role

    @property
    def name(self) -> str:
        return self._name

    async def new_token(self, *, expiration: datetime = None) \
            -> LoginTokenData:
        token_id = uuid.uuid4().hex
        token_key = token_bytes()
        self.login.tokens[token_id] = (self.name, token_key)
        return LoginTokenData(self.name, token_id, token_key)

    @asynccontextmanager
    async def new_session(self) -> AsyncIterator[Session]:
        identity = self.name
        config = self.config
        _ = await self.get()
        mailbox_set, filter_set = config.set_cache.get(identity, (None, None))
        if not mailbox_set or not filter_set:
            mailbox_set = MailboxSet()
            filter_set = FilterSet()
            if config.demo_data and identity == config.demo_user:
                await self._load_demo(config.demo_data_resource,
                                      mailbox_set, filter_set)
            config.set_cache[identity] = (mailbox_set, filter_set)
        yield Session(identity, config, mailbox_set, filter_set)

    async def _load_demo(self, resource: str, mailbox_set: MailboxSet,
                         filter_set: FilterSet) -> None:
        inbox = await mailbox_set.get_mailbox('INBOX')
        await self._load_demo_mailbox(resource, 'INBOX', inbox)
        mbx_names = sorted(resource_listdir(resource, 'demo'))
        for name in mbx_names:
            if name == 'sieve':
                await self._load_demo_sieve(resource, name, filter_set)
            elif name != 'INBOX':
                await mailbox_set.add_mailbox(name)
                mbx = await mailbox_set.get_mailbox(name)
                await self._load_demo_mailbox(resource, name, mbx)

    async def _load_demo_sieve(self, resource: str, name: str,
                               filter_set: FilterSet) -> None:
        path = os.path.join('demo', name)
        with closing(resource_stream(resource, path)) as sieve_stream:
            sieve = sieve_stream.read()
        await filter_set.put('demo', sieve)
        await filter_set.set_active('demo')

    async def _load_demo_mailbox(self, resource: str, name: str,
                                 mbx: MailboxData) -> None:
        path = os.path.join('demo', name)
        msg_names = sorted(resource_listdir(resource, path))
        for msg_name in msg_names:
            if msg_name == '.readonly':
                mbx._readonly = True
                continue
            elif msg_name.startswith('.'):
                continue
            msg_path = os.path.join(path, msg_name)
            with closing(resource_stream(resource, msg_path)) as msg_stream:
                flags_line = msg_stream.readline()
                msg_timestamp = float(msg_stream.readline())
                msg_data = msg_stream.read()
            msg_dt = datetime.fromtimestamp(msg_timestamp, timezone.utc)
            msg_flags = {Flag(flag) for flag in flags_line.split()}
            if Recent in msg_flags:
                msg_flags.remove(Recent)
                msg_recent = True
            else:
                msg_recent = False
            append_msg = AppendMessage(msg_data, msg_dt, frozenset(msg_flags))
            await mbx.append(append_msg, recent=msg_recent)

    async def get(self) -> UserMetadata:
        data = self.login.users.get(self.name)
        if data is None:
            raise UserNotFound(self.name)
        return data

    async def set(self, data: UserMetadata) -> None:
        if self._role != 'admin' and data.role:
            raise NotAllowedError('Cannot assign role.')
        self.login.users[self.name] = data

    async def delete(self) -> None:
        try:
            del self.login.users[self.name]
        except KeyError as exc:
            raise UserNotFound(self.name) from exc
        self.config.set_cache.pop(self.name, None)
        for token_id, (name, _) in list(self.login.tokens.items()):
            if name == self.name:
                del self.login.tokens[token_id]



"""Re-exposes the :func:`re.compile` function. The typing of the arguments and
returned :class:`~typing.Pattern` are replaced such that only
:class:`bytes`-based patterns are allowed and either :class:`bytes` or
:class:`memoryview` may be used for the strings.

See Also:
    :mod:`re`

"""

import re as _re

__all__ = ['compile']


compile = _re.compile



"""Contains the context variables that are made available by the IMAP server to
backend implementations.

See Also:
    :mod:`contextvars`

Attributes:
    subsystem: The :class:`~pymap.concurrent.Subsystem` for concurrency
        primitives.
    current_command: The currently executing
        :class:`~pymap.parsing.command.Command`.
    socket_info: :class:`~proxyprotocol.sock.SocketInfo` about the currently
        connected client.
    language_code: The language code string, e.g. ``en``.
    connection_exit: The active :class:`~contextlib.AsyncExitStack` that will
        be exited when the connection closes.

"""

from __future__ import annotations

from contextlib import AsyncExitStack
from contextvars import ContextVar

from proxyprotocol.sock import SocketInfo

from .concurrent import Subsystem
from .parsing.command import Command

__all__ = ['subsystem', 'current_command', 'socket_info', 'language_code',
           'connection_exit']

subsystem: ContextVar[Subsystem] = ContextVar(
    'subsystem', default=Subsystem.for_asyncio())
current_command: ContextVar[Command] = ContextVar('current_command')
socket_info: ContextVar[SocketInfo] = ContextVar('socket_info')
language_code: ContextVar[str] = ContextVar('language_code')
connection_exit: ContextVar[AsyncExitStack] = ContextVar('connection_exit')




from __future__ import annotations

import asyncio
import binascii
import logging
import re
import sys
from argparse import ArgumentParser
from asyncio import shield, StreamReader, StreamWriter, AbstractServer, \
    CancelledError, TimeoutError
from base64 import b64encode, b64decode
from contextlib import closing, AsyncExitStack
from typing import TypeVar, Union, Optional, Iterable, List, Awaitable
from uuid import uuid4

from proxyprotocol import ProxyProtocolResult
from proxyprotocol.sock import SocketInfo
from proxyprotocol.version import ProxyProtocolVersion
from pymap.concurrent import Event
from pymap.config import IMAPConfig
from pymap.context import subsystem, current_command, socket_info, \
    connection_exit
from pymap.exceptions import ResponseError
from pymap.interfaces.backend import ServiceInterface
from pymap.interfaces.login import LoginInterface
from pymap.parsing.command import Command
from pymap.parsing.commands import Commands
from pymap.parsing.command.nonauth import AuthenticateCommand, StartTLSCommand
from pymap.parsing.command.select import IdleCommand
from pymap.parsing.response import ResponseContinuation, Response, \
    ResponseCode, ResponseBad, ResponseNo, ResponseBye, ResponseOk, \
    CommandResponse
from pymap.parsing.state import ParsingState, ParsingInterrupt, \
    ExpectContinuation
from pymap.sockets import InheritedSockets
from pysasl import ServerChallenge, ChallengeResponse, AuthenticationError, \
    AuthenticationCredentials

from .state import ConnectionState

__all__ = ['IMAPService', 'IMAPServer', 'IMAPConnection']

_Ret = TypeVar('_Ret')
_log = logging.getLogger(__name__)


class IMAPService(ServiceInterface):  # pragma: no cover
    """A pymap service implementing an IMAP server."""

    @classmethod
    def add_arguments(cls, parser: ArgumentParser) -> None:
        group = parser.add_argument_group('imap service')
        group.add_argument('--host', metavar='IFACE', action='append',
                           help='the network interface to listen on')
        group.add_argument('--port', metavar='NUM', default='143',
                           help='the port or service name to listen on')
        group.add_argument('--cert', metavar='FILE', help='cert file for TLS')
        group.add_argument('--key', metavar='FILE', help='key file for TLS')
        if InheritedSockets.supports('systemd'):
            group.add_argument('--systemd-sockets', action='store_const',
                               dest='inherited_sockets', const='systemd',
                               help='use systemd inherited sockets')
        else:
            parser.set_defaults(inherited_sockets=None)
        group.add_argument('--no-tls', dest='tls', action='store_false',
                           help='disable TLS')
        pp_choices = [v.name.lower() for v in ProxyProtocolVersion]
        group.add_argument('--proxy-protocol', choices=pp_choices,
                           help='the PROXY protocol version string')

    async def start(self) -> Awaitable:
        backend = self.backend
        config = self.config
        servers: List[AbstractServer] = []
        imap_server = IMAPServer(backend.login, config)
        if config.args.inherited_sockets:
            sockets = InheritedSockets.of(config.args.inherited_sockets).get()
            if not sockets:
                raise ValueError('No inherited sockets found')
            for sock in sockets:
                server = await asyncio.start_server(imap_server, sock=sock)
                servers.append(server)
        else:
            server = await asyncio.start_server(
                imap_server, host=config.host, port=config.port)
            servers.append(server)
        return asyncio.gather(*[self._run(server) for server in servers])

    @classmethod
    async def _run(cls, server: AbstractServer) -> None:
        async with server:
            await server.serve_forever()


class IMAPServer:
    """Callable object that creates and runs :class:`IMAPConnection` objects
    when :func:`asyncio.start_server` receives a new connection.

    Args:
        login: Login callback that takes authentication credentials and returns
            a :class:`~pymap.interfaces.session.SessionInterface` object.
        config: Settings to use for the IMAP server.

    """

    __slots__ = ['commands', '_login', '_config']

    def __init__(self, login: LoginInterface, config: IMAPConfig) -> None:
        super().__init__()
        self.commands = config.commands
        self._login = login
        self._config = config

    async def __call__(self, reader: StreamReader,
                       writer: StreamWriter) -> None:
        conn = IMAPConnection(self.commands, self._config, reader, writer)
        state = ConnectionState(self._login, self._config)
        async with AsyncExitStack() as stack:
            connection_exit.set(stack)
            stack.enter_context(closing(conn))
            await conn.run(state)


class IMAPConnection:
    """Runs a single IMAP connection from start to finish.

    Args:
        commands: Defines the IMAP commands available to the connection.
        config: Settings to use for the IMAP connection.
        reader: The input stream for the socket.
        writer: The output stream for the socket.

    """

    _lines = re.compile(r'\r?\n')
    _literal_plus = re.compile(br'{(\d+)\+}\r?\n$')

    __slots__ = ['commands', 'config', 'params', 'bad_command_limit',
                 'reader', 'writer', 'pp_result']

    def __init__(self, commands: Commands, config: IMAPConfig,
                 reader: StreamReader,
                 writer: StreamWriter) -> None:
        super().__init__()
        self.commands = commands
        self.config = config
        self.params = config.parsing_params
        self.bad_command_limit = config.bad_command_limit
        self.pp_result: Optional[ProxyProtocolResult] = None
        self._reset_streams(reader, writer)

    def _reset_streams(self, reader: StreamReader,
                       writer: StreamWriter) -> None:
        self.reader = reader
        self.writer = writer
        socket_info.set(SocketInfo(writer, self.pp_result,
                                   unique_id=uuid4().bytes))

    async def _read_proxy_protocol(self) -> None:
        self.pp_result = await self.config.proxy_protocol.read(self.reader)
        self._reset_streams(self.reader, self.writer)

    def close(self) -> None:
        self.writer.close()

    @classmethod
    def _print(cls, log_format: str, output: Union[str, bytes]) -> None:
        if _log.isEnabledFor(logging.DEBUG):
            uid = socket_info.get().unique_id.hex()
            if not isinstance(output, str):
                output = str(output, 'utf-8', 'replace')
            lines = cls._lines.split(output)
            if not lines[-1]:
                lines = lines[:-1]
            for line in lines:
                _log.debug(log_format, uid, line)

    def _exec(self, future: Awaitable[_Ret]) -> Awaitable[_Ret]:
        return subsystem.get().execute(future)

    async def readline(self) -> memoryview:
        buf = bytearray(await self.reader.readline())
        while True:
            if not buf.endswith(b'\n'):
                raise EOFError()
            elif buf.endswith(b'+}\n') or buf.endswith(b'+}\r\n'):
                lit_plus = self._literal_plus.search(buf)
            else:
                lit_plus = None
            if lit_plus:
                literal_length = int(lit_plus.group(1))
                buf += await self.reader.readexactly(literal_length)
                buf += await self.reader.readline()
            else:
                self._print('%s -->| %s', buf)
                return memoryview(buf)

    async def read_continuation(self, literal_length: int) -> memoryview:
        extra_literal = await self.reader.readexactly(literal_length)
        self._print('%s -->| %s', extra_literal)
        extra_line = await self.readline()
        extra = extra_literal + bytes(extra_line)
        return memoryview(extra)

    async def authenticate(self, state: ConnectionState, mech_name: bytes) \
            -> Optional[AuthenticationCredentials]:
        mech = state.auth.get_server(mech_name)
        if not mech:
            return None
        responses: List[ChallengeResponse] = []
        while True:
            try:
                creds, final = mech.server_attempt(responses)
            except ServerChallenge as chal:
                chal_bytes = b64encode(chal.data)
                cont = ResponseContinuation(chal_bytes)
                await self.write_response(cont)
                resp_bytes = bytes(await self.read_continuation(0))
                if resp_bytes.rstrip(b'\r\n') == b'*':
                    raise AuthenticationError('Authentication canceled.')
                try:
                    resp_dec = b64decode(resp_bytes)
                except binascii.Error as exc:
                    raise AuthenticationError() from exc
                else:
                    responses.append(ChallengeResponse(chal.data, resp_dec))
            else:
                if final is not None:
                    cont = ResponseContinuation(b64encode(final))
                    await self.write_response(cont)
                    await self.read_continuation(0)
                return creds

    async def _interrupt(self, state: ConnectionState,
                         interrupt: ParsingInterrupt,
                         continuations: List[memoryview]) -> None:
        expected = interrupt.expected
        if isinstance(expected, ExpectContinuation):
            cont = ResponseContinuation(expected.message)
            await self.write_response(cont)
            ret = await self.read_continuation(expected.literal_length)
            continuations.append(ret)
        else:
            raise TypeError(expected) from interrupt

    async def read_command(self, state: ConnectionState) -> Command:
        line = await self.readline()
        conts: List[memoryview] = []
        while True:
            parsing_state = ParsingState(continuations=conts)
            params = self.params.copy(parsing_state)
            try:
                cmd, _ = self.commands.parse(line, params)
            except ParsingInterrupt as interrupt:
                await self._interrupt(state, interrupt, conts)
            else:
                return cmd

    async def read_idle_done(self, cmd: IdleCommand) -> bool:
        buf = await self.read_continuation(0)
        ok, _ = cmd.parse_done(buf)
        return ok

    async def write_response(self, resp: Response) -> None:
        await resp.async_write(self.writer)
        try:
            await self.writer.drain()
        except ConnectionError:
            pass
        else:
            self._print('%s <--| %s', bytes(resp))

    async def start_tls(self) -> None:
        loop = asyncio.get_event_loop()
        transport = self.writer.transport
        protocol = transport.get_protocol()
        ssl_context = self.config.ssl_context
        new_transport = await loop.start_tls(
            transport, protocol, ssl_context, server_side=True)
        new_protocol = new_transport.get_protocol()
        new_writer = StreamWriter(new_transport, new_protocol,
                                  self.reader, loop)
        self._reset_streams(self.reader, new_writer)
        self._print('%s <->| %s', b'<TLS handshake>')

    async def send_error_disconnect(self) -> None:
        _, exc, _ = sys.exc_info()
        if isinstance(exc, CancelledError):
            resp = ResponseBye(b'Server has closed the connection.',
                               ResponseCode.of(b'UNAVAILABLE'))
        else:
            resp = ResponseBye(b'Unhandled server error.',
                               ResponseCode.of(b'SERVERBUG'))
        try:
            await self.write_response(resp)
        except IOError:
            pass

    async def write_updates(self, untagged: Iterable[Response]) -> None:
        for i, resp in enumerate(untagged):
            await self.write_response(resp)

    async def handle_updates(self, state: ConnectionState, done: Event,
                             cmd: IdleCommand) -> None:
        while not done.is_set():
            untagged = await self._exec(state.receive_updates(cmd, done))
            await shield(self.write_updates(untagged))

    async def idle(self, state: ConnectionState, cmd: IdleCommand) \
            -> CommandResponse:
        response = await self._exec(state.do_command(cmd))
        if not isinstance(response, ResponseOk):
            return response
        await self.write_response(ResponseContinuation(b'Idling.'))
        done = subsystem.get().new_event()
        updates_task = asyncio.create_task(
            self.handle_updates(state, done, cmd))
        done_task = asyncio.create_task(self.read_idle_done(cmd))
        updates_exc: Optional[Exception] = None
        done_exc: Optional[Exception] = None
        try:
            ok = await done_task
        except Exception as exc:
            done_exc = exc
        finally:
            done.set()
        try:
            await updates_task
        except Exception as exc:
            updates_exc = exc
        if updates_exc:
            raise updates_exc
        elif done_exc:
            raise done_exc
        elif not ok:
            return ResponseBad(cmd.tag, b'Expected "DONE".')
        else:
            return response

    async def run(self, state: ConnectionState) -> None:
        """Start the socket communication with the IMAP greeting, and then
        enter the command/response cycle.

        Args:
            state: Defines the interaction with the backend plugin.

        """
        await self._read_proxy_protocol()
        self._print('%s +++| %s', str(socket_info.get()))
        bad_commands = 0
        try:
            greeting = await self._exec(state.do_greeting())
        except ResponseError as exc:
            resp = exc.get_response(b'*')
            resp.condition = ResponseBye.condition
            await self.write_response(resp)
            return
        else:
            await self.write_response(greeting)
        while True:
            try:
                cmd = await self.read_command(state)
            except (ConnectionError, EOFError):
                break
            except CancelledError:
                await self.send_error_disconnect()
                break
            except Exception:
                await self.send_error_disconnect()
                raise
            else:
                prev_cmd = current_command.set(cmd)
                try:
                    if isinstance(cmd, AuthenticateCommand):
                        creds = await self.authenticate(state, cmd.mech_name)
                        response = await self._exec(
                            state.do_authenticate(cmd, creds))
                    elif isinstance(cmd, IdleCommand):
                        response = await self.idle(state, cmd)
                    else:
                        response = await self._exec(state.do_command(cmd))
                except ResponseError as exc:
                    resp = exc.get_response(cmd.tag)
                    await self.write_response(resp)
                    if resp.is_terminal:
                        break
                except AuthenticationError as exc:
                    msg = bytes(str(exc), 'utf-8', 'surrogateescape')
                    resp = ResponseBad(cmd.tag, msg)
                    await self.write_response(resp)
                except TimeoutError:
                    resp = ResponseNo(cmd.tag, b'Operation timed out.',
                                      ResponseCode.of(b'TIMEOUT'))
                    await self.write_response(resp)
                except CancelledError:
                    await self.send_error_disconnect()
                    break
                except Exception:
                    await self.send_error_disconnect()
                    raise
                else:
                    await self.write_response(response)
                    if response.is_bad:
                        bad_commands += 1
                        if self.bad_command_limit \
                                and bad_commands >= self.bad_command_limit:
                            msg = b'Too many errors, disconnecting.'
                            response.add_untagged(ResponseBye(msg))
                    else:
                        bad_commands = 0
                    if response.is_terminal:
                        break
                    if isinstance(cmd, StartTLSCommand) \
                            and isinstance(response, ResponseOk):
                        await self.start_tls()
                finally:
                    await state.do_cleanup()
                    current_command.reset(prev_cmd)
        self._print('%s ---| %s', b'<disconnected>')




from __future__ import annotations

from typing import Any, Tuple, Sequence, Mapping, Optional
from typing_extensions import Final

from aioredis import Redis  # type: ignore

from . import ScriptBase
from ..keys import CleanupKeys, NamespaceKeys, ContentKeys, MailboxKeys

__all__ = ['MailboxScripts']


class MailboxScripts:

    def __init__(self) -> None:
        super().__init__()
        self.add: Final = MessageAdd()
        self.copy: Final = MessageCopy()
        self.move: Final = MessageMove()
        self.update: Final = MessageUpdate()
        self.delete: Final = MessageDelete()
        self.snapshot: Final = MailboxSnapshot()


class MessageAdd(ScriptBase[Tuple[int, bytes, bytes]]):

    def __init__(self) -> None:
        super().__init__('message_add')

    def _convert(self, ret: Tuple[bytes, bytes, bytes]) \
            -> Tuple[int, bytes, bytes]:
        return (int(ret[0]), ret[1], ret[2])

    async def __call__(self, redis: Redis, ns_keys: NamespaceKeys,
                       ct_keys: ContentKeys, mbx_keys: MailboxKeys, *,
                       recent: bool, date: bytes, flags: Sequence[str],
                       email_id: bytes, thread_id: bytes,
                       thread_keys: Sequence[str],
                       message: bytes, message_json: Mapping[str, Any],
                       header: bytes, header_json: Mapping[str, Any]) \
            -> Tuple[int, bytes, bytes]:
        keys = [mbx_keys.max_uid, mbx_keys.uids, mbx_keys.seq,
                mbx_keys.content, mbx_keys.changes, mbx_keys.recent,
                mbx_keys.deleted, mbx_keys.unseen,
                ns_keys.max_modseq, ns_keys.thread_keys,
                ns_keys.content_refs, ct_keys.data]
        return await self.eval(redis, keys, [
            int(recent), self._pack(flags), date,
            email_id, thread_id, self._pack(thread_keys),
            message, self._pack(message_json),
            header, self._pack(header_json)])


class MessageCopy(ScriptBase[int]):

    def __init__(self) -> None:
        super().__init__('message_copy')

    def _convert(self, ret: Tuple[bytes]) -> int:
        return int(ret[0])

    async def __call__(self, redis: Redis, ns_keys: NamespaceKeys,
                       mbx_keys: MailboxKeys, dest_mbx_keys: MailboxKeys, *,
                       source_uid: int, recent: bool) -> int:
        keys = [mbx_keys.uids, dest_mbx_keys.max_uid, dest_mbx_keys.uids,
                dest_mbx_keys.seq, dest_mbx_keys.content,
                dest_mbx_keys.changes, dest_mbx_keys.recent,
                dest_mbx_keys.deleted, dest_mbx_keys.unseen,
                ns_keys.max_modseq, ns_keys.content_refs]
        return await self.eval(redis, keys, [
            source_uid, int(recent)])


class MessageMove(ScriptBase[int]):

    def __init__(self) -> None:
        super().__init__('message_move')

    def _convert(self, ret: Tuple[bytes]) -> int:
        return int(ret[0])

    async def __call__(self, redis: Redis, ns_keys: NamespaceKeys,
                       mbx_keys: MailboxKeys, dest_mbx_keys: MailboxKeys, *,
                       source_uid: int, recent: bool) -> int:
        keys = [mbx_keys.uids, mbx_keys.seq, mbx_keys.content,
                mbx_keys.changes, mbx_keys.recent, mbx_keys.deleted,
                mbx_keys.unseen, dest_mbx_keys.max_uid, dest_mbx_keys.uids,
                dest_mbx_keys.seq, dest_mbx_keys.content,
                dest_mbx_keys.changes, dest_mbx_keys.recent,
                dest_mbx_keys.deleted, dest_mbx_keys.unseen,
                ns_keys.max_modseq]
        return await self.eval(redis, keys, [
            source_uid, int(recent)])


class MessageUpdate(ScriptBase[bytes]):

    def __init__(self) -> None:
        super().__init__('message_update')

    async def __call__(self, redis: Redis,
                       ns_keys: NamespaceKeys, mbx_keys: MailboxKeys, *,
                       uid: int, flags: Sequence[str], mode: bytes) -> bytes:
        keys = [mbx_keys.uids, mbx_keys.changes, mbx_keys.deleted,
                mbx_keys.unseen, ns_keys.max_modseq]
        return await self.eval(redis, keys, [
            uid, mode, self._pack(flags)])


class MessageDelete(ScriptBase[None]):

    def __init__(self) -> None:
        super().__init__('message_delete')

    async def __call__(self, redis: Redis, ns_keys: NamespaceKeys,
                       mbx_keys: MailboxKeys, cl_keys: CleanupKeys, *,
                       uids: Sequence[int]) -> None:
        keys = [mbx_keys.uids, mbx_keys.seq, mbx_keys.content,
                mbx_keys.changes, mbx_keys.recent, mbx_keys.deleted,
                mbx_keys.unseen, ns_keys.max_modseq, cl_keys.contents]
        return await self.eval(redis, keys, [
            self._pack(uids),
            mbx_keys.root.named['namespace'],
            mbx_keys.root.named['mailbox_id']])


class MailboxSnapshot(ScriptBase[Tuple[int, int, int, int, Optional[int]]]):

    def __init__(self) -> None:
        super().__init__('mailbox_snapshot')

    def _convert(self, ret: Tuple[bytes, bytes, bytes, bytes, bytes]) \
            -> Tuple[int, int, int, int, Optional[int]]:
        return (int(ret[0]), int(ret[1]), int(ret[2]),
                int(ret[3]), self._maybe_int(ret[4]))

    async def __call__(self, redis: Redis, mbx_keys: MailboxKeys) \
            -> Tuple[int, int, int, int, Optional[int]]:
        keys = [mbx_keys.max_uid, mbx_keys.uids, mbx_keys.seq,
                mbx_keys.recent, mbx_keys.unseen]
        return await self.eval(redis, keys, [])




from __future__ import annotations

from abc import abstractmethod
from datetime import datetime
from typing import Optional, NamedTuple, AsyncContextManager
from typing_extensions import Protocol

from pysasl import AuthenticationCredentials

from .session import SessionInterface
from ..user import UserMetadata

__all__ = ['LoginTokenData', 'LoginInterface', 'IdentityInterface']


class LoginTokenData(NamedTuple):
    """Data required to build or verify a login token.

    Args:
        identity: The token identity, or the username to authorize as.
        identifier: The token identifier, used to determine *key*.
        key: The token key, used to securely hash and verify the token.

    """

    identity: str
    identifier: str
    key: bytes


class LoginInterface(Protocol):
    """Defines the authentication operations for backends."""

    @abstractmethod
    async def authenticate(self, credentials: AuthenticationCredentials) \
            -> IdentityInterface:
        """Authenticate and authorize the credentials.

        Raises:
            :exc:`~pymap.exceptions.InvalidAuth`

        """
        ...


class IdentityInterface(Protocol):
    """Defines the operations available to a user identity that has been
    authenticated and authorized. This user identity may or may not "exist" in
    the backend.

    """

    @property
    def name(self) -> str:
        """The SASL authorization identity of the logged-in user."""
        ...

    @abstractmethod
    def new_session(self) -> AsyncContextManager[SessionInterface]:
        """Authenticate and authorize the credentials, returning a new IMAP
        session.

        Args:
            credentials: Authentication credentials supplied by the user.

        Raises:
            :class:`~pymap.exceptions.UserNotFound`

        """
        ...

    @abstractmethod
    async def new_token(self, *, expiration: datetime = None) \
            -> Optional[LoginTokenData]:
        """Authenticate and authorize the credentials, returning an identifier
        and private key that can be used to create and verify tokens.

        Since tokens should use their own private key, backends may return
        ``None`` if it does not support tokens or the user does not have a
        private key.

        Args:
            expiration: When the token should stop being valid.

        Raises:
            :exc:`~pymap.exceptions.UserNotFound`

        """
        ...

    @abstractmethod
    async def get(self) -> UserMetadata:
        """Return the metadata associated with the user identity.

        Raises:
            :exc:`~pymap.exceptions.UserNotFound`

        """
        ...

    @abstractmethod
    async def set(self, metadata: UserMetadata) -> None:
        """Assign new metadata to the user identity.

        Args:
            metadata: New metadata, such as password.

        """
        ...

    @abstractmethod
    async def delete(self) -> None:
        """Delete existing metadata for the user identity.

        Raises:
            :exc:`~pymap.exceptions.UserNotFound`

        """
        ...




from __future__ import annotations

from abc import abstractmethod
from typing import Any, Tuple, Optional, FrozenSet, Iterable, Sequence
from typing_extensions import Protocol

from .filter import FilterSetInterface
from .message import MessageInterface
from .mailbox import MailboxInterface
from ..concurrent import Event
from ..flags import FlagOp
from ..parsing.message import AppendMessage
from ..parsing.specials import SequenceSet, Flag, SearchKey, ObjectId
from ..parsing.response.code import AppendUid, CopyUid
from ..selected import SelectedMailbox

__all__ = ['SessionInterface']


class SessionInterface(Protocol):
    """Corresponds to a single, authenticated IMAP session."""

    __slots__: Sequence[str] = []

    @property
    @abstractmethod
    def owner(self) -> str:
        """The SASL authorization identity of the logged-in user."""
        ...

    @property
    @abstractmethod
    def filter_set(self) -> Optional[FilterSetInterface[Any]]:
        """Manages the active and inactive filters for the login user."""
        ...

    @abstractmethod
    def close(self) -> None:
        """Called at the end of a session, to clean up any resources."""
        ...

    @abstractmethod
    async def cleanup(self) -> None:
        """Called after every operation, success or failure, to allow the
        backend to clean up resources.

        Note:
            Any exception raised by this method will be silently ignored.

        """
        ...

    @abstractmethod
    async def list_mailboxes(self, ref_name: str, filter_: str,
                             subscribed: bool = False,
                             selected: SelectedMailbox = None) \
            -> Tuple[Iterable[Tuple[str, Optional[str], Sequence[bytes]]],
                     Optional[SelectedMailbox]]:
        """List the mailboxes owned by the user.

        See Also:
            `RFC 3501 6.3.8.
            <https://tools.ietf.org/html/rfc3501#section-6.3.8>`_,
            `RFC 3501 6.3.9.
            <https://tools.ietf.org/html/rfc3501#section-6.3.9>`_

        Args:
            ref_name: Mailbox reference name.
            filter_: Mailbox name with possible wildcards.
            subscribed: If True, only list the subscribed mailboxes.
            selected: If applicable, the currently selected mailbox name.

        """
        ...

    @abstractmethod
    async def get_mailbox(self, name: str, selected: SelectedMailbox = None) \
            -> Tuple[MailboxInterface, Optional[SelectedMailbox]]:
        """Retrieves a :class:`~pymap.interfaces.mailbox.MailboxInterface`
        object corresponding to an existing mailbox owned by the user. Raises
        an exception if the mailbox does not yet exist.

        Args:
            name: The name of the mailbox.
            selected: If applicable, the currently selected mailbox name.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`

        """
        ...

    @abstractmethod
    async def create_mailbox(self, name: str,
                             selected: SelectedMailbox = None) \
            -> Tuple[ObjectId, Optional[SelectedMailbox]]:
        """Creates a new mailbox owned by the user.

        See Also:
            `RFC 3501 6.3.3.
            <https://tools.ietf.org/html/rfc3501#section-6.3.3>`_

        Args:
            name: The name of the mailbox.
            selected: If applicable, the currently selected mailbox name.

        Raises:
            :class:`~pymap.exceptions.MailboxConflict`

        """
        ...

    @abstractmethod
    async def delete_mailbox(self, name: str,
                             selected: SelectedMailbox = None) \
            -> Optional[SelectedMailbox]:
        """Deletes the mailbox owned by the user.

        See Also:
            `RFC 3501 6.3.4.
            <https://tools.ietf.org/html/rfc3501#section-6.3.4>`_

        Args:
            name: The name of the mailbox.
            selected: If applicable, the currently selected mailbox name.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`
            :class:`~pymap.exceptions.MailboxHasChildren`

        """
        ...

    @abstractmethod
    async def rename_mailbox(self, before_name: str, after_name: str,
                             selected: SelectedMailbox = None) \
            -> Optional[SelectedMailbox]:
        """Renames the mailbox owned by the user.

        See Also:
            `RFC 3501 6.3.5.
            <https://tools.ietf.org/html/rfc3501#section-6.3.5>`_

        Args:
            before_name: The name of the mailbox before the rename.
            after_name: The name of the mailbox after the rename.
            selected: If applicable, the currently selected mailbox name.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`
            :class:`~pymap.exceptions.MailboxConflict`

        """
        ...

    @abstractmethod
    async def subscribe(self, name: str, selected: SelectedMailbox = None) \
            -> Optional[SelectedMailbox]:
        """Mark the given folder name as subscribed, whether or not the given
        folder name currently exists.

        See Also:
            `RFC 3501 6.3.6.
            <https://tools.ietf.org/html/rfc3501#section-6.3.6>`_

        Args:
            name: The name of the mailbox.
            selected: If applicable, the currently selected mailbox name.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`

        """
        ...

    @abstractmethod
    async def unsubscribe(self, name: str, selected: SelectedMailbox = None) \
            -> Optional[SelectedMailbox]:
        """Remove the given folder name from the subscription list, whether or
        not the given folder name currently exists.

        See Also:
            `RFC 3501 6.3.6.
            <https://tools.ietf.org/html/rfc3501#section-6.3.6>`_

        Args:
            name: The name of the mailbox.
            selected: If applicable, the currently selected mailbox name.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`

        """
        ...

    @abstractmethod
    async def append_messages(self, name: str,
                              messages: Sequence[AppendMessage],
                              selected: SelectedMailbox = None) \
            -> Tuple[AppendUid, Optional[SelectedMailbox]]:
        """Appends a message to the end of the mailbox.

        See Also:
            `RFC 3502 6.3.11.
            <https://tools.ietf.org/html/rfc3502#section-6.3.11>`_

        Args:
            name: The name of the mailbox.
            messages: The messages to append.
            selected: If applicable, the currently selected mailbox name.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`
            :class:`~pymap.exceptions.AppendFailure`

        """
        ...

    @abstractmethod
    async def select_mailbox(self, name: str, readonly: bool = False) \
            -> Tuple[MailboxInterface, SelectedMailbox]:
        """Selects an existing mailbox owned by the user. The returned session
        is then used as the ``selected`` argument to other methods to fetch
        mailbox updates.

        See Also:
            `RFC 3501 6.3.1.
            <https://tools.ietf.org/html/rfc3501#section-6.3.1>`_,
            `RFC 3501 6.3.2.
            <https://tools.ietf.org/html/rfc3501#section-6.3.2>`_

        Args:
            name: The name of the mailbox.
            readonly: True if the mailbox is read-only.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`

        """
        ...

    @abstractmethod
    async def check_mailbox(self, selected: SelectedMailbox, *,
                            wait_on: Event = None,
                            housekeeping: bool = False) -> SelectedMailbox:
        """Checks for any updates in the mailbox.

        If ``wait_on`` is given, this method should block until either this
        event is signalled or the mailbox has detected updates. This method may
        be called continuously as long as ``wait_on`` is not signalled.

        If ``housekeeping`` is True, perform any house-keeping necessary by the
        mailbox backend, which may be a slower operation.

        See Also:
            `RFC 3501 6.1.2.
            <https://tools.ietf.org/html/rfc3501#section-6.1.2>`_,
            `RFC 3501 6.4.1.
            <https://tools.ietf.org/html/rfc3501#section-6.4.1>`_
            `RFC 2177 <https://tools.ietf.org/html/rfc2177>`_

        Args:
            selected: The selected mailbox session.
            wait_on: If given, block until this event signals.
            housekeeping: If True, the backend may perform additional
                housekeeping operations if necessary.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`

        """
        ...

    @abstractmethod
    async def fetch_messages(self, selected: SelectedMailbox,
                             sequence_set: SequenceSet, set_seen: bool) \
            -> Tuple[Iterable[Tuple[int, MessageInterface]], SelectedMailbox]:
        """Get a list of loaded message objects corresponding to given sequence
        set.

        Args:
            selected: The selected mailbox session.
            sequence_set: Sequence set of message sequences or UIDs.
            set_seen: True if the messages should get the ``\\Seen`` flag.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`

        """
        ...

    @abstractmethod
    async def search_mailbox(self, selected: SelectedMailbox,
                             keys: FrozenSet[SearchKey]) \
            -> Tuple[Iterable[Tuple[int, MessageInterface]], SelectedMailbox]:
        """Get the messages in the current mailbox that meet all of the
        given search criteria.

        See Also:
            `RFC 3501 7.2.5.
            <https://tools.ietf.org/html/rfc3501#section-7.2.5>`_

        Args:
            selected: The selected mailbox session.
            keys: Search keys specifying the message criteria.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`

        """
        ...

    @abstractmethod
    async def expunge_mailbox(self, selected: SelectedMailbox,
                              uid_set: SequenceSet = None) -> SelectedMailbox:
        """All messages that are marked as deleted are immediately expunged
        from the mailbox.

        See Also:
            `RFC 3501 6.4.3.
            <https://tools.ietf.org/html/rfc3501#section-6.4.3>`_
            `RFC 4315 2.1 <https://tools.ietf.org/html/rfc4315#section-2.1>`_

        Args:
            selected: The selected mailbox session.
            uid_set: Only the messages in the given UID set should be expunged.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`
            :class:`~pymap.exceptions.MailboxReadOnly`

        """
        ...

    @abstractmethod
    async def copy_messages(self, selected: SelectedMailbox,
                            sequence_set: SequenceSet,
                            mailbox: str) \
            -> Tuple[Optional[CopyUid], SelectedMailbox]:
        """Copy a set of messages into the given mailbox.

        See Also:
            `RFC 3501 6.4.7.
            <https://tools.ietf.org/html/rfc3501#section-6.4.7>`_

        Args:
            selected: The selected mailbox session.
            sequence_set: Sequence set of message sequences or UIDs.
            mailbox: Name of the mailbox to copy messages into.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`
            :class:`~pymap.exceptions.MailboxReadOnly`

        """
        ...

    @abstractmethod
    async def move_messages(self, selected: SelectedMailbox,
                            sequence_set: SequenceSet,
                            mailbox: str) \
            -> Tuple[Optional[CopyUid], SelectedMailbox]:
        """Move a set of messages into the given mailbox, removing them from
        the selected mailbox.

        See Also:
            `RFC 6851 <https://tools.ietf.org/html/rfc6851>`_

        Args:
            selected: The selected mailbox session.
            sequence_set: Sequence set of message sequences or UIDs.
            mailbox: Name of the mailbox to copy messages into.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`
            :class:`~pymap.exceptions.MailboxReadOnly`

        """
        ...

    @abstractmethod
    async def update_flags(self, selected: SelectedMailbox,
                           sequence_set: SequenceSet,
                           flag_set: FrozenSet[Flag],
                           mode: FlagOp = FlagOp.REPLACE) \
            -> Tuple[Iterable[Tuple[int, MessageInterface]], SelectedMailbox]:
        """Update the flags for the given set of messages.

        See Also:
            `RFC 3501 6.4.6.
            <https://tools.ietf.org/html/rfc3501#section-6.4.6>`_

        Args:
            selected: The selected mailbox session.
            sequence_set: Sequence set of message sequences or UIDs.
            flag_set: Set of flags to update.
            mode: Update mode for the flag set.

        Raises:
            :class:`~pymap.exceptions.MailboxNotFound`
            :class:`~pymap.exceptions.MailboxReadOnly`

        """
        ...



"""Module containing the general exceptions that may be used by pymap."""

from __future__ import annotations

from abc import abstractmethod, ABCMeta
from typing import Optional

from .parsing.specials import SearchKey
from .parsing.response import Response, ResponseCode, ResponseNo, ResponseOk, \
    ResponseBye

__all__ = ['ResponseError', 'CloseConnection', 'NotSupportedError',
           'TemporaryFailure', 'SearchNotAllowed', 'InvalidAuth',
           'AuthorizationFailure', 'NotAllowedError', 'IncompatibleData',
           'MailboxError', 'MailboxNotFound', 'MailboxConflict',
           'MailboxHasChildren', 'MailboxReadOnly', 'AppendFailure',
           'UserNotFound']


class ResponseError(Exception, metaclass=ABCMeta):
    """The base exception for all custom errors that are used to generate a
    response to an IMAP command.

    """

    @abstractmethod
    def get_response(self, tag: bytes) -> Response:
        """Build an IMAP response for the error.

        Args:
            tag: The command tag that generated the error.

        """
        ...


class CloseConnection(ResponseError):
    """Raised when the connection should be closed immediately after sending
    the provided response.

    """

    def get_response(self, tag: bytes) -> ResponseOk:
        response = ResponseOk(tag, b'Logout successful.')
        response.add_untagged(ResponseBye(b'Logging out.'))
        return response


class NotSupportedError(ResponseError, NotImplementedError):
    """Raised when an action is taken that might be syntactically valid, but is
    not supported by the server or backend.

    """

    def __init__(self, msg: str = 'Operation not supported.') -> None:
        super().__init__(msg)
        self._raw = msg.encode('utf-8')

    def get_response(self, tag: bytes) -> ResponseNo:
        return ResponseNo(tag, self._raw, ResponseCode.of(b'CANNOT'))


class SearchNotAllowed(NotSupportedError):
    """The ``SEARCH`` command contained a search key that could not be
    executed by the mailbox.

    Args:
        key: The search key that failed.

    """

    def __init__(self, key: SearchKey) -> None:
        super().__init__(f'SEARCH {key.value_str} not supported.')


class InvalidAuth(ResponseError):
    """The ``LOGIN`` or ``AUTHENTICATE`` commands received credentials that the
    IMAP backend has rejected.

    """

    def __init__(self, msg: str = 'Invalid authentication credentials.') \
            -> None:
        super().__init__(msg)
        self._raw = msg.encode('utf-8')

    def get_response(self, tag: bytes) -> ResponseNo:
        return ResponseNo(tag, self._raw,
                          ResponseCode.of(b'AUTHENTICATIONFAILED'))


class AuthorizationFailure(InvalidAuth):
    """The credentials in ``LOGIN`` or ``AUTHENTICATE`` were authenticated but
    failed to authorize as the requested identity.

    """

    def __init__(self, msg: str = 'Authorization failed.') -> None:
        super().__init__(msg)

    def get_response(self, tag: bytes) -> ResponseNo:
        return ResponseNo(tag, self._raw,
                          ResponseCode.of(b'AUTHORIZATIONFAILED'))


class NotAllowedError(ResponseError):
    """The operation is not allowed due to access controls."""

    def __init__(self, msg: str = 'Operation not allowed.') -> None:
        super().__init__(msg)
        self._raw = msg.encode('utf-8')

    def get_response(self, tag: bytes) -> ResponseNo:
        return ResponseNo(tag, self._raw, ResponseCode.of(b'NOPERM'))


class IncompatibleData(InvalidAuth):
    """The ``LOGIN`` or ``AUTHENTICATE`` command could not succeed because the
    detected mailbox data was not in a compatible format.

    """

    def __init__(self, msg: str = 'Incompatible mailbox data.') -> None:
        super().__init__(msg)


class TemporaryFailure(ResponseError):
    """The operation failed, but may succeed if tried again. The ``[INUSE]``
    response code is added to the response.

    """

    def __init__(self, msg: str) -> None:
        super().__init__(msg)
        self._raw = msg.encode('utf-8')

    def get_response(self, tag: bytes) -> ResponseNo:
        return ResponseNo(tag, self._raw, ResponseCode.of(b'INUSE'))


class MailboxError(ResponseError):
    """Parent exception for errors related to a mailbox.

    Args:
        mailbox: The name of the mailbox.
        message: The response message for the error.
        code: Optional response code for the error.

    """

    def __init__(self, mailbox: Optional[str], message: bytes,
                 code: Optional[ResponseCode] = None) -> None:
        super().__init__()
        self.mailbox = mailbox
        self.message = message
        self.code = code

    def get_response(self, tag: bytes) -> ResponseNo:
        return ResponseNo(tag, self.message, self.code)


class MailboxNotFound(MailboxError):
    """The requested mailbox was not found.

    Args:
        mailbox: The name of the mailbox
        try_create: True if creating the mailbox first may help.

    """

    def __init__(self, mailbox: str, *, try_create: bool = False) -> None:
        code = ResponseCode.of(b'TRYCREATE' if try_create else b'NONEXISTENT')
        super().__init__(mailbox, b'Mailbox does not exist.', code)


class MailboxConflict(MailboxError):
    """The mailbox cannot be created or renamed because of a naming conflict
    with another mailbox.

    Args:
        mailbox: The name of the mailbox.

    """

    def __init__(self, mailbox: str) -> None:
        super().__init__(mailbox, b'Mailbox already exists.',
                         ResponseCode.of(b'ALREADYEXISTS'))


class MailboxHasChildren(MailboxError):
    """The mailbox cannot be deleted because there are other inferior
    hierarchical mailboxes below it.

    Args:
        mailbox: The name of the mailbox.

    """

    def __init__(self, mailbox: str) -> None:
        super().__init__(mailbox, b'Mailbox has inferior hierarchical names.')


class MailboxReadOnly(MailboxError):
    """The mailbox is opened read-only and the requested operation is not
    allowed.

    Args:
        mailbox: The name of the mailbox.

    """

    def __init__(self, mailbox: Optional[str] = None) -> None:
        super().__init__(mailbox, b'Mailbox is read-only.',
                         ResponseCode.of(b'READ-ONLY'))


class AppendFailure(MailboxError):
    """The mailbox append operation failed."""


class UserNotFound(ResponseError):
    """The requested user was not found."""

    def get_response(self, tag: bytes) -> ResponseNo:
        return ResponseNo(tag, b'User not found.',
                          ResponseCode.of(b'NONEXISTENT'))




from __future__ import annotations

import errno
import os
import os.path
from datetime import datetime
from mailbox import Maildir as _Maildir, MaildirMessage
from typing import Dict, Optional, Union, Tuple, FrozenSet, Iterable, \
    AsyncIterable
from typing_extensions import Final, Literal

from pymap.concurrent import Event, ReadWriteLock
from pymap.context import subsystem
from pymap.exceptions import MailboxHasChildren, NotSupportedError
from pymap.flags import FlagOp
from pymap.interfaces.message import CachedMessage
from pymap.listtree import ListTree
from pymap.mailbox import MailboxSnapshot
from pymap.message import BaseMessage, BaseLoadedMessage
from pymap.mime import MessageContent
from pymap.parsing.message import AppendMessage
from pymap.parsing.specials import ObjectId, FetchRequirement
from pymap.parsing.specials.flag import Flag, Seen
from pymap.selected import SelectedSet, SelectedMailbox

from .flags import MaildirFlags
from .io import NoChanges
from .layout import MaildirLayout
from .subscriptions import Subscriptions
from .uidlist import Record, UidList
from ..mailbox import MailboxDataInterface, MailboxSetInterface

__all__ = ['Maildir', 'Message', 'MailboxData', 'MailboxSet']


class Maildir(_Maildir):

    @property
    def _path_new(self) -> str:
        return self._paths['new']  # type: ignore

    @property
    def _path_cur(self) -> str:
        return self._paths['cur']  # type: ignore

    def _join(self, subpath: str) -> str:
        base_path: str = self._path  # type: ignore
        return os.path.join(base_path, subpath)

    def _split(self, subpath: str) \
            -> Tuple[Union[Literal['new'], Literal['cur']], str]:
        subdir, name = os.path.split(subpath)
        if subdir == 'new':
            return 'new', name
        elif subdir == 'cur':
            return 'cur', name
        else:
            raise ValueError(subdir)

    def _lookup(self, key: str) -> str:
        return super()._lookup(key)  # type: ignore

    def _update(self, key: str, subpath: str) -> None:
        self._toc[key] = subpath  # type: ignore

    def claim_new(self) -> Iterable[str]:
        """Checks for messages in the ``new`` subdirectory, moving them to
        ``cur`` and returning their keys.

        """
        new_subdir = self._path_new
        cur_subdir = self._path_cur
        for name in os.listdir(new_subdir):
            new_path = os.path.join(new_subdir, name)
            cur_path = os.path.join(cur_subdir, name)
            try:
                os.rename(new_path, cur_path)
            except FileNotFoundError:
                pass
            else:
                yield name.rsplit(self.colon, 1)[0]

    def move_message(self, key: str, dest: Maildir, dest_subdir: str) -> str:
        """Moves the message to another maildir."""
        subpath = self._lookup(key)
        subdir, name = self._split(subpath)
        dest_subpath = os.path.join(dest_subdir, name)
        path = self._join(subpath)
        dest_path = dest._join(dest_subpath)
        os.rename(path, dest_path)
        return name

    def get_message_metadata(self, key: str) -> MaildirMessage:
        """Like :meth:`~mailbox.Maildir.get_message` but the message contents
        are not read from disk.

        """
        msg = MaildirMessage()
        subpath = self._lookup(key)
        subdir, name = self._split(subpath)
        msg.set_subdir(subdir)
        if self.colon in name:
            msg.set_info(name.rsplit(self.colon, 1)[-1])
        msg.set_date(os.path.getmtime(self._join(subpath)))
        return msg

    def update_metadata(self, key: str, msg: MaildirMessage) -> None:
        """Uses :func:`os.rename` to atomically update the message filename
        based on :meth:`~mailbox.MaildirMessage.get_info`.

        """
        subpath = self._lookup(key)
        subdir, name = self._split(subpath)
        new_subdir = msg.get_subdir()
        new_name = key + self.colon + msg.get_info()
        if subdir != new_subdir:
            raise ValueError('Message subdir may not be updated')
        elif name != new_name:
            new_subpath = os.path.join(msg.get_subdir(), new_name)
            old_path = self._join(subpath)
            new_path = self._join(new_subpath)
            os.rename(old_path, new_path)
            self._update(key, new_subpath)


class Message(BaseMessage):

    __slots__ = ['recent', '_maildir', '_key']

    def __init__(self, uid: int, internal_date: datetime,
                 permanent_flags: Iterable[Flag], *, expunged: bool = False,
                 email_id: ObjectId = None, thread_id: ObjectId = None,
                 recent: bool = False, maildir: Maildir = None,
                 key: str = None) -> None:
        super().__init__(uid, internal_date, permanent_flags,
                         expunged=expunged, email_id=email_id,
                         thread_id=thread_id)
        self.recent: Final = recent
        self._maildir = maildir
        self._key = key

    async def load_content(self, requirement: FetchRequirement) \
            -> LoadedMessage:
        if self._key is None or self._maildir is None \
                or requirement.has_none(FetchRequirement.CONTENT):
            return LoadedMessage(self, requirement, None)
        try:
            maildir_msg = self._maildir.get_message(self._key)
        except (KeyError, FileNotFoundError):
            return LoadedMessage(self, requirement, None)
        else:
            content = MessageContent.parse(bytes(maildir_msg))
            return LoadedMessage(self, requirement, content)

    @classmethod
    def copy_expunged(cls, msg: Message) -> Message:
        return cls(msg.uid, msg.internal_date, msg.permanent_flags,
                   expunged=True, email_id=msg.email_id,
                   thread_id=msg.thread_id, maildir=msg._maildir, key=msg._key)

    @classmethod
    def to_maildir(cls, append_msg: AppendMessage, recent: bool,
                   maildir_flags: MaildirFlags) -> MaildirMessage:
        flag_str = maildir_flags.to_maildir(append_msg.flag_set)
        when = append_msg.when or datetime.now()
        maildir_msg = MaildirMessage(append_msg.literal)
        maildir_msg.set_flags(flag_str)
        maildir_msg.set_subdir('new' if recent else 'cur')
        maildir_msg.set_date(when.timestamp())
        return maildir_msg

    @classmethod
    def from_maildir(cls, uid: int, maildir_msg: MaildirMessage,
                     maildir: Maildir, key: str,
                     email_id: Optional[ObjectId],
                     thread_id: Optional[ObjectId],
                     maildir_flags: MaildirFlags) -> Message:
        flag_set = maildir_flags.from_maildir(maildir_msg.get_flags())
        recent = maildir_msg.get_subdir() == 'new'
        msg_dt = datetime.fromtimestamp(maildir_msg.get_date())
        return cls(uid, msg_dt, flag_set,
                   email_id=email_id, thread_id=thread_id,
                   recent=recent, maildir=maildir, key=key)


class LoadedMessage(BaseLoadedMessage):
    pass


class MailboxData(MailboxDataInterface[Message]):

    def __init__(self, mailbox_id: ObjectId, maildir: Maildir,
                 path: str) -> None:
        super().__init__()
        self._mailbox_id = mailbox_id
        self._maildir = maildir
        self._path = path
        self._uid_validity = 0
        self._next_uid = 0
        self._flags: Optional[MaildirFlags] = None
        self._messages_lock = subsystem.get().new_rwlock()
        self._selected_set = SelectedSet()

    @classmethod
    def _get_object_id(cls, rec: Record, field: str) -> Optional[ObjectId]:
        return ObjectId.maybe(rec.fields.get(field))

    @property
    def mailbox_id(self) -> ObjectId:
        return self._mailbox_id

    @property
    def readonly(self) -> bool:
        return False

    @property
    def uid_validity(self) -> int:
        return self._uid_validity

    @property
    def maildir_flags(self) -> MaildirFlags:
        if self._flags is not None:
            return self._flags
        self._flags = flags = MaildirFlags.file_read(self._path)
        return flags

    @property
    def permanent_flags(self) -> FrozenSet[Flag]:
        return self.maildir_flags.permanent_flags

    @property
    def messages_lock(self) -> ReadWriteLock:
        return self._messages_lock

    @property
    def selected_set(self) -> SelectedSet:
        return self._selected_set

    async def _get_maildir_msg(self, uid: int) \
            -> Tuple[Record, MaildirMessage]:
        async with UidList.with_read(self._path) as uidl:
            record = uidl.get(uid)
        maildir = self._maildir
        key = record.key
        async with self.messages_lock.read_lock():
            maildir_msg = maildir.get_message_metadata(key)
        return record, maildir_msg

    async def update_selected(self, selected: SelectedMailbox, *,
                              wait_on: Event = None) -> SelectedMailbox:
        if wait_on is not None:
            await wait_on.wait(timeout=1.0)
        all_messages = [msg async for msg in self.messages()]
        selected.set_messages(all_messages)
        return selected

    async def append(self, append_msg: AppendMessage, *,
                     recent: bool = False) -> Message:
        maildir = self._maildir
        email_id = ObjectId.random_email_id()
        thread_id = ObjectId.random_thread_id()
        async with self.messages_lock.write_lock():
            maildir_msg = Message.to_maildir(append_msg, recent,
                                             self.maildir_flags)
            key = maildir.add(maildir_msg)
            filename = key + ':' + maildir_msg.get_info()
        async with UidList.with_write(self._path) as uidl:
            fields = {'E': str(email_id), 'T': str(thread_id)}
            new_rec = Record(uidl.next_uid, fields, filename)
            uidl.next_uid += 1
            uidl.set(new_rec)
        return Message.from_maildir(
            new_rec.uid, maildir_msg, maildir, key, email_id, thread_id,
            self.maildir_flags)

    async def copy(self, uid: int, destination: MailboxData, *,
                   recent: bool = False) -> Optional[int]:
        dest_maildir = destination._maildir
        try:
            record, maildir_msg = await self._get_maildir_msg(uid)
        except KeyError:
            return None
        copy_msg = MaildirMessage(maildir_msg)
        copy_msg.set_subdir('new' if recent else 'cur')
        async with destination.messages_lock.write_lock():
            dest_key = dest_maildir.add(copy_msg)
            dest_filename = dest_key + ':' + copy_msg.get_info()
        async with UidList.with_write(destination._path) as uidl:
            new_rec = Record(uidl.next_uid, record.fields, dest_filename)
            uidl.next_uid += 1
            uidl.set(new_rec)
        return new_rec.uid

    async def move(self, uid: int, destination: MailboxData, *,
                   recent: bool = False) -> Optional[int]:
        maildir = self._maildir
        dest_maildir = destination._maildir
        async with UidList.with_read(self._path) as uidl:
            try:
                rec = uidl.get(uid)
            except KeyError:
                return None
        dest_subdir = 'new' if recent else 'cur'
        async with destination.messages_lock.write_lock(), \
                self.messages_lock.write_lock():
            try:
                new_filename = maildir.move_message(
                    rec.key, dest_maildir, dest_subdir)
            except (KeyError, FileNotFoundError):
                return None
        async with UidList.with_write(destination._path) as uidl:
            new_rec = Record(uidl.next_uid, rec.fields, new_filename)
            uidl.next_uid += 1
            uidl.set(new_rec)
        return new_rec.uid

    async def get(self, uid: int, cached_msg: CachedMessage) -> Message:
        maildir = self._maildir
        try:
            record, maildir_msg = await self._get_maildir_msg(uid)
        except (KeyError, FileNotFoundError):
            if not isinstance(cached_msg, Message):
                raise TypeError(cached_msg)
            return Message.copy_expunged(cached_msg)
        key = record.key
        email_id = self._get_object_id(record, 'E')
        thread_id = self._get_object_id(record, 'T')
        return Message.from_maildir(
            uid, maildir_msg, maildir, key, email_id, thread_id,
            self.maildir_flags)

    async def update(self, uid: int, cached_msg: CachedMessage,
                     flag_set: FrozenSet[Flag], mode: FlagOp) -> Message:
        maildir = self._maildir
        try:
            record, maildir_msg = await self._get_maildir_msg(uid)
        except (KeyError, FileNotFoundError):
            if not isinstance(cached_msg, Message):
                raise TypeError(cached_msg)
            msg = Message.copy_expunged(cached_msg)
            msg.permanent_flags = mode.apply(msg.permanent_flags, flag_set)
            return msg
        key = record.key
        email_id = self._get_object_id(record, 'E')
        thread_id = self._get_object_id(record, 'T')
        existing_flags = self.maildir_flags.from_maildir(
            maildir_msg.get_flags())
        new_flags = mode.apply(existing_flags, flag_set)
        new_flags_str = self.maildir_flags.to_maildir(new_flags)
        maildir_msg.set_flags(new_flags_str)
        try:
            maildir.update_metadata(key, maildir_msg)
        except (KeyError, FileNotFoundError):
            pass
        return Message.from_maildir(
            uid, maildir_msg, maildir, key, email_id, thread_id,
            self.maildir_flags)

    async def delete(self, uids: Iterable[int]) -> None:
        async with UidList.with_read(self._path) as uidl:
            records = uidl.get_all(uids)
        async with self.messages_lock.write_lock():
            for uid, rec in records.items():
                try:
                    self._maildir.remove(rec.key)
                except (KeyError, FileNotFoundError):
                    pass

    async def claim_recent(self, selected: SelectedMailbox) -> None:
        async with self.messages_lock.write_lock():
            keys = self._maildir.claim_new()
        async with UidList.with_read(self._path) as uidl:
            for rec in uidl.records:
                if rec.key in keys:
                    selected.session_flags.add_recent(rec.uid)

    async def cleanup(self) -> None:
        self._maildir.clean()
        keys = await self._get_keys()
        async with UidList.with_write(self._path) as uidl:
            for rec in list(uidl.records):
                key = rec.key
                info = keys.get(key)
                if info is None:
                    uidl.remove(rec.uid)
                else:
                    filename = key + ':' + info
                    new_rec = Record(rec.uid, rec.fields, filename)
                    uidl.set(new_rec)

    async def messages(self) -> AsyncIterable[Message]:
        async with UidList.with_read(self._path) as uidl:
            uids = {rec.uid: rec for rec in uidl.records}
        maildir = self._maildir
        async with self.messages_lock.read_lock():
            for uid, rec in uids.items():
                email_id = self._get_object_id(rec, 'E')
                thread_id = self._get_object_id(rec, 'T')
                try:
                    maildir_msg = maildir.get_message_metadata(rec.key)
                except (KeyError, FileNotFoundError):
                    pass
                else:
                    yield Message.from_maildir(
                        uid, maildir_msg, maildir, rec.key,
                        email_id, thread_id, self.maildir_flags)

    async def reset(self) -> MailboxData:
        keys = await self._get_keys()
        async with UidList.with_write(self._path) as uidl:
            for rec in uidl.records:
                keys.pop(rec.key, None)
            if not keys:
                raise NoChanges()
            for key, info in keys.items():
                filename = key + ':' + info
                fields = {'E': str(ObjectId.random_email_id()),
                          'T': str(ObjectId.random_thread_id())}
                new_rec = Record(uidl.next_uid, fields, filename)
                uidl.next_uid += 1
                uidl.set(new_rec)
        self._uid_validity = uidl.uid_validity
        self._next_uid = uidl.next_uid
        return self

    async def snapshot(self) -> MailboxSnapshot:
        exists = 0
        recent = 0
        unseen = 0
        first_unseen: Optional[int] = None
        next_uid = self._next_uid
        async for msg in self.messages():
            exists += 1
            if msg.recent:
                recent += 1
            if Seen not in msg.permanent_flags:
                unseen += 1
                if first_unseen is None:
                    first_unseen = exists
        return MailboxSnapshot(self.mailbox_id, self.readonly,
                               self.uid_validity, self.permanent_flags,
                               self.session_flags, exists, recent, unseen,
                               first_unseen, next_uid)

    async def _get_keys(self) -> Dict[str, str]:
        keys: Dict[str, str] = {}
        async with self.messages_lock.read_lock():
            for key in self._maildir.keys():
                try:
                    msg = self._maildir.get_message_metadata(key)
                except (KeyError, FileNotFoundError):
                    pass
                else:
                    keys[key] = msg.get_info()
        return keys


class MailboxSet(MailboxSetInterface[MailboxData]):

    def __init__(self, maildir: Maildir, layout: MaildirLayout) -> None:
        super().__init__()
        self._layout = layout
        self._inbox_maildir = maildir
        self._path = layout.path
        self._cache: Dict[str, MailboxData] = {}

    @property
    def delimiter(self) -> str:
        return '/'

    async def set_subscribed(self, name: str, subscribed: bool) -> None:
        async with Subscriptions.with_write(self._path) as subs:
            subs.set(name, subscribed)

    async def list_subscribed(self) -> ListTree:
        async with Subscriptions.with_read(self._path) as subs:
            subscribed = frozenset(subs.subscribed)
        mailboxes = [name for name in self._layout.list_folders(self.delimiter)
                     if name in subscribed]
        return ListTree(self.delimiter).update('INBOX', *mailboxes)

    async def list_mailboxes(self) -> ListTree:
        mailboxes = self._layout.list_folders(self.delimiter)
        return ListTree(self.delimiter).update('INBOX', *mailboxes)

    async def get_mailbox(self, name: str) -> MailboxData:
        if name == 'INBOX':
            maildir = self._inbox_maildir
        else:
            try:
                maildir = self._layout.get_folder(name, self.delimiter)
            except FileNotFoundError as exc:
                raise KeyError(name) from exc
        if name in self._cache:
            mbx = self._cache[name]
        else:
            path = self._layout.get_path(name, self.delimiter)
            async with UidList.with_open(path) as uidl:
                mailbox_id = ObjectId(uidl.global_uid)
            mbx = MailboxData(mailbox_id, maildir, path)
            self._cache[name] = mbx
        return await mbx.reset()

    async def add_mailbox(self, name: str) -> ObjectId:
        try:
            self._layout.add_folder(name, self.delimiter)
        except FileExistsError as exc:
            raise KeyError(name) from exc
        path = self._layout.get_path(name, self.delimiter)
        async with UidList.with_open(path) as uidl:
            global_uid = uidl.global_uid
        return ObjectId(global_uid)

    async def delete_mailbox(self, name: str) -> None:
        try:
            self._layout.remove_folder(name, self.delimiter)
        except FileNotFoundError as exc:
            raise KeyError(name) from exc
        except OSError as exc:
            if exc.errno == errno.ENOTEMPTY:
                raise MailboxHasChildren(name) from exc
            raise exc

    async def rename_mailbox(self, before: str, after: str) -> None:
        if before == 'INBOX':
            raise NotSupportedError()  # TODO
        else:
            self._layout.rename_folder(before, after, self.delimiter)




import unittest

from pymap.listtree import ListEntry, ListTree


class TestListEntry(unittest.TestCase):

    def test_attributes(self) -> None:
        self.assertEqual([b'HasNoChildren'],
                         ListEntry('', True, None, False).attributes)
        self.assertEqual([b'HasChildren'],
                         ListEntry('', True, None, True).attributes)
        self.assertEqual([b'Noselect', b'HasChildren'],
                         ListEntry('', False, None, True).attributes)
        self.assertEqual([b'HasNoChildren', b'Marked'],
                         ListEntry('', True, True, False).attributes)
        self.assertEqual([b'HasNoChildren', b'Unmarked'],
                         ListEntry('', True, False, False).attributes)


class TestListTree(unittest.TestCase):

    def setUp(self) -> None:
        tree = ListTree('/')
        tree.update('INBOX',
                    'Sent',
                    'Trash',
                    'Important/One',
                    'Important/Two/Three',
                    'To Do',
                    'To Do/Quickly')
        tree.set_marked('To Do', marked=True)
        tree.set_marked('To Do/Quickly', unmarked=True)
        self.tree = tree

    def test_list(self) -> None:
        self.assertEqual([ListEntry('INBOX', True, None, False),
                          ListEntry('Sent', True, None, False),
                          ListEntry('Trash', True, None, False),
                          ListEntry('Important', False, None, True),
                          ListEntry('Important/One', True, None, False),
                          ListEntry('Important/Two', False, None, True),
                          ListEntry('Important/Two/Three', True, None, False),
                          ListEntry('To Do', True, True, True),
                          ListEntry('To Do/Quickly', True, False, False)],
                         list(self.tree.list()))

    def test_list_matching(self) -> None:
        self.assertEqual([ListEntry('Important', False, None, True),
                          ListEntry('Important/One', True, None, False),
                          ListEntry('Important/Two', False, None, True),
                          ListEntry('Important/Two/Three', True, None, False)],
                         list(self.tree.list_matching('Important', '*')))
        self.assertEqual([ListEntry('Important/One', True, None, False),
                          ListEntry('Important/Two', False, None, True)],
                         list(self.tree.list_matching('Important/', '%')))
        self.assertEqual([ListEntry('INBOX', True, None, False)],
                         list(self.tree.list_matching('inbox', '')))

    def test_get_renames(self) -> None:
        self.assertEqual([], self.tree.get_renames('Missing', 'Test'))
        self.assertEqual([('Important/One', 'Trivial/One'),
                          ('Important/Two/Three', 'Trivial/Two/Three')],
                         self.tree.get_renames('Important', 'Trivial'))
        self.assertEqual([('Important/Two/Three', 'Trivial/Two/Three')],
                         self.tree.get_renames('Important/Two', 'Trivial/Two'))




from __future__ import annotations

from datetime import datetime
from typing import Optional, Iterable

import msgpack  # type: ignore
from aioredis import Redis  # type: ignore

from pymap.message import BaseMessage, BaseLoadedMessage
from pymap.mime import MessageContent, MessageHeader, MessageBody
from pymap.parsing.specials import Flag, ObjectId, FetchRequirement

from .keys import NamespaceKeys, ContentKeys

__all__ = ['Message', 'LoadedMessage']


class Message(BaseMessage):

    __slots__ = ['_redis', '_ns_keys']

    def __init__(self, uid: int, internal_date: datetime,
                 permanent_flags: Iterable[Flag], *, expunged: bool = False,
                 email_id: ObjectId = None, thread_id: ObjectId = None,
                 redis: Redis = None, ns_keys: NamespaceKeys = None) -> None:
        super().__init__(uid, internal_date, permanent_flags,
                         expunged=expunged, email_id=email_id,
                         thread_id=thread_id)
        self._redis = redis
        self._ns_keys = ns_keys

    async def _load_full(self, redis: Redis, ct_keys: ContentKeys) \
            -> MessageContent:
        literal, full_json_raw = await redis.hmget(
            ct_keys.data, b'full', b'full-json')
        if literal is None or full_json_raw is None:
            raise ValueError(f'Missing message content: {self.email_id}')
        full_json = msgpack.unpackb(full_json_raw, raw=False)
        return MessageContent.from_json(literal, full_json)

    async def _load_header(self, redis: Redis, ct_keys: ContentKeys) \
            -> MessageContent:
        literal, header_json_raw = await redis.hmget(
            ct_keys.data, b'header', b'header-json')
        if literal is None or header_json_raw is None:
            raise ValueError(f'Missing message header: {self.email_id}')
        header_json = msgpack.unpackb(header_json_raw, raw=False)
        header = MessageHeader.from_json(literal, header_json)
        body = MessageBody.empty()
        return MessageContent(literal, header, body)

    async def load_content(self, requirement: FetchRequirement) \
            -> LoadedMessage:
        redis = self._redis
        ns_keys = self._ns_keys
        if redis is None or ns_keys is None:
            return LoadedMessage(self, requirement, None)
        ct_keys = ContentKeys(ns_keys, self.email_id)
        content: Optional[MessageContent] = None
        if requirement & FetchRequirement.BODY:
            content = await self._load_full(redis, ct_keys)
        elif requirement & FetchRequirement.HEADER:
            content = await self._load_header(redis, ct_keys)
        return LoadedMessage(self, requirement, content)

    @classmethod
    def copy_expunged(cls, msg: Message) -> Message:
        return cls(msg.uid, msg.internal_date, msg.permanent_flags,
                   expunged=True, email_id=msg.email_id,
                   thread_id=msg.thread_id, redis=msg._redis,
                   ns_keys=msg._ns_keys)


class LoadedMessage(BaseLoadedMessage):
    pass




from __future__ import annotations

import re
from abc import abstractmethod, ABCMeta
from email.headerregistry import Address, AddressHeader
from typing import Sequence, List, Pattern

from pymap.mime import MessageContent
from pymap.parsing.message import AppendMessage
from sievelib.commands import Command  # type: ignore

from .util import AddressPart, MatchType, SizeComparator, str_list

__all__ = ['SieveTest']


class SieveTest(metaclass=ABCMeta):
    """Base class for all sieve tests.

    See Also:
        `RFC 5228 5. <https://tools.ietf.org/html/rfc5228#section-5>`_

    """

    @classmethod
    def of(cls, cmd: Command) -> SieveTest:
        """Given a test command, determine the implementation sub-class and
        return an instance.

        Args:
            cmd: The test command.

        """
        if cmd.name == 'allof':
            return AllOfTest([SieveTest.of(test)
                              for test in cmd.arguments['tests']])
        elif cmd.name == 'anyof':
            return AnyOfTest([SieveTest.of(test)
                              for test in cmd.arguments['tests']])
        elif cmd.name == 'not':
            return NotTest(SieveTest.of(cmd.arguments['test']))
        elif cmd.name == 'true':
            return TrueTest()
        elif cmd.name == 'false':
            return FalseTest()
        elif cmd.name in ('address', 'envelope'):
            address_part = AddressPart.of(cmd.arguments.get('address-part'))
            match_type = MatchType.of(cmd.arguments.get('match-type'))
            case_sensitive = cmd.arguments.get('comparator') == '"i;octet"'
            header_list = str_list(cmd.arguments['header-list'])
            key_list = str_list(cmd.arguments['key-list'])
            if cmd.name == 'address':
                return AddressTest(header_list, key_list, address_part,
                                   match_type, case_sensitive)
            else:
                return EnvelopeTest(header_list, key_list, address_part,
                                    match_type, case_sensitive)
        elif cmd.name == 'exists':
            header_names = str_list(cmd.arguments['header-names'])
            return ExistsTest(header_names)
        elif cmd.name == 'header':
            match_type = MatchType.of(cmd.arguments.get('match-type'))
            case_sensitive = cmd.arguments.get('comparator') == '"i;octet"'
            header_names = str_list(cmd.arguments['header-names'])
            key_list = str_list(cmd.arguments['key-list'])
            return HeaderTest(header_names, key_list, match_type,
                              case_sensitive)
        elif cmd.name == 'size':
            comparator = SizeComparator.of(cmd.arguments['comparator'])
            limit = int(cmd.arguments['limit'])
            return SizeTest(comparator, limit)
        else:
            raise NotImplementedError(cmd.name)

    @abstractmethod
    def run(self, sender: str, recipient: str, append_msg: AppendMessage,
            content: MessageContent) -> bool:
        """Run the test implementation on the appended message data, returning
        True if the test matched the data. Must be overridden by sub-classes.

        Args:
            sender: The envelope sender of the message.
            recipient: The envelope recipient of the message.
            append_msg: The message to be appended.
            content: The parsed message content.

        """
        ...


class AllOfTest(SieveTest):

    def __init__(self, tests: Sequence[SieveTest]) -> None:
        super().__init__()
        self._tests = tests

    def run(self, sender: str, recipient: str, append_msg: AppendMessage,
            content: MessageContent) -> bool:
        return all(test.run(sender, recipient, append_msg, content)
                   for test in self._tests)


class AnyOfTest(SieveTest):

    def __init__(self, tests: Sequence[SieveTest]) -> None:
        super().__init__()
        self._tests = tests

    def run(self, sender: str, recipient: str, append_msg: AppendMessage,
            content: MessageContent) -> bool:
        return any(test.run(sender, recipient, append_msg, content)
                   for test in self._tests)


class NotTest(SieveTest):

    def __init__(self, test: SieveTest) -> None:
        super().__init__()
        self._test = test

    def run(self, sender: str, recipient: str, append_msg: AppendMessage,
            content: MessageContent) -> bool:
        return not self._test.run(sender, recipient, append_msg, content)


class TrueTest(SieveTest):

    def run(self, sender: str, recipient: str, append_msg: AppendMessage,
            content: MessageContent) -> bool:
        return True


class FalseTest(SieveTest):

    def run(self, sender: str, recipient: str, append_msg: AppendMessage,
            content: MessageContent) -> bool:
        return False


class MatchTest(SieveTest):

    _wildcards = re.compile(r'([\*\?])')

    @classmethod
    def _compile(cls, key_list: Sequence[str], match_type: MatchType,
                 case_sensitive: bool) -> Sequence[Pattern]:
        flags = 0 if case_sensitive else re.I
        ret: List[Pattern] = []
        for key in key_list:
            if match_type == MatchType.MATCHES:
                pattern_parts: List[str] = []
                for part in cls._wildcards.split(key):
                    if part == '*':
                        pattern_parts.append('.*?')
                    elif part == '?':
                        pattern_parts.append('.')
                    else:
                        pattern_parts.append(re.escape(part))
                pattern = ''.join(pattern_parts)
            else:
                pattern = re.escape(key)
            if match_type in (MatchType.IS, MatchType.MATCHES):
                pattern = '^' + pattern + '$'
            ret.append(re.compile(pattern, flags))
        return ret


class AddressTest(MatchTest):

    def __init__(self, header_list: Sequence[str], key_list: Sequence[str],
                 address_part: AddressPart, match_type: MatchType,
                 case_sensitive: bool) -> None:
        super().__init__()
        self._header_set = {hdr.encode('ascii') for hdr in header_list}
        self._address_part = address_part
        self._patterns = self._compile(key_list, match_type, case_sensitive)

    def run(self, sender: str, recipient: str, append_msg: AppendMessage,
            content: MessageContent) -> bool:
        part = self._address_part
        for name in self._header_set:
            values = content.header.parsed.get(name, [])
            for value in values:
                if isinstance(value, AddressHeader):
                    for address in value.addresses:
                        if part == AddressPart.LOCALPART:
                            addr_part = address.username
                        elif part == AddressPart.DOMAIN:
                            addr_part = address.domain
                        else:
                            addr_part = address.addr_spec
                        for pattern in self._patterns:
                            if pattern.match(addr_part) is not None:
                                return True
        return False


class EnvelopeTest(AddressTest):

    def _check_address(self, envelope_part: bytes, address: str) -> bool:
        if envelope_part not in self._header_set:
            return False
        try:
            addr_obj = Address(addr_spec=address)
        except Exception:
            addr_obj = Address(username=address)
        part = self._address_part
        if part == AddressPart.LOCALPART:
            addr_part = addr_obj.username
        elif part == AddressPart.DOMAIN:
            addr_part = addr_obj.domain
        else:
            addr_part = addr_obj.addr_spec
            if addr_part == '<>':
                addr_part = ''
        for pattern in self._patterns:
            if pattern.match(addr_part) is not None:
                return True
        return False

    def run(self, sender: str, recipient: str, append_msg: AppendMessage,
            content: MessageContent) -> bool:
        if self._check_address(b'from', sender):
            return True
        elif self._check_address(b'to', recipient):
            return True
        else:
            return False


class ExistsTest(SieveTest):

    def __init__(self, header_names: Sequence[str]) -> None:
        super().__init__()
        self._header_names = {hdr.encode('ascii') for hdr in header_names}

    def run(self, sender: str, recipient: str, append_msg: AppendMessage,
            content: MessageContent) -> bool:
        return all(name in content.header.parsed
                   for name in self._header_names)


class HeaderTest(MatchTest):

    def __init__(self, header_names: Sequence[str], key_list: Sequence[str],
                 match_type: MatchType, case_sensitive: bool) -> None:
        super().__init__()
        self._header_names = [hdr.encode('ascii') for hdr in header_names]
        self._patterns = self._compile(key_list, match_type, case_sensitive)

    def run(self, sender: str, recipient: str, append_msg: AppendMessage,
            content: MessageContent) -> bool:
        for name in self._header_names:
            values = content.header.parsed.get(name, [])
            for value in values:
                value_str = str(value)
                for pattern in self._patterns:
                    if pattern.match(value_str) is not None:
                        return True
        return False


class SizeTest(SieveTest):

    def __init__(self, comparator: SizeComparator, limit: int) -> None:
        super().__init__()
        self._comparator = comparator
        self._limit = limit

    def run(self, sender: str, recipient: str, append_msg: AppendMessage,
            content: MessageContent) -> bool:
        length = len(content)
        if self._comparator == SizeComparator.OVER:
            return length > self._limit
        else:
            return length < self._limit




import unittest

from pymap.parsing.response import CommandResponse, ResponseContinuation, \
    ResponseBad, ResponseNo, ResponseOk, ResponseBye, ResponseCode


_alert = ResponseCode.of(b'ALERT')


class TestResponse(unittest.TestCase):

    def test_bytes(self):
        resp = CommandResponse(b'tag', b'response test')
        self.assertEqual(b'tag response test\r\n', bytes(resp))
        self.assertEqual(b'tag response test\r\n', bytes(resp))
        resp.add_untagged_ok(b'test data 1')
        self.assertEqual(b'* OK test data 1\r\n'
                         b'tag response test\r\n', bytes(resp))
        resp.add_untagged_ok(b'test data 2')
        self.assertEqual(b'* OK test data 1\r\n'
                         b'* OK test data 2\r\n'
                         b'tag response test\r\n', bytes(resp))


class TestResponseContinuation(unittest.TestCase):

    def test_bytes(self):
        resp = ResponseContinuation(b'test continuation')
        self.assertEqual(b'+ test continuation\r\n', bytes(resp))


class TestResponseBad(unittest.TestCase):

    def test_bytes(self):
        resp1 = ResponseBad(b'tag', b'bad response')
        self.assertEqual(b'tag BAD bad response\r\n', bytes(resp1))
        resp2 = ResponseBad(b'tag', b'bad response', _alert)
        self.assertEqual(b'tag BAD [ALERT] bad response\r\n', bytes(resp2))


class TestResponseNo(unittest.TestCase):

    def test_bytes(self):
        resp1 = ResponseNo(b'tag', b'invalid response')
        self.assertEqual(b'tag NO invalid response\r\n', bytes(resp1))
        resp2 = ResponseNo(b'tag', b'invalid response', _alert)
        self.assertEqual(b'tag NO [ALERT] invalid response\r\n', bytes(resp2))


class TestResponseOk(unittest.TestCase):

    def test_bytes(self):
        resp1 = ResponseOk(b'tag', b'ok response')
        self.assertEqual(b'tag OK ok response\r\n', bytes(resp1))
        resp2 = ResponseOk(b'tag', b'ok response', _alert)
        self.assertEqual(b'tag OK [ALERT] ok response\r\n', bytes(resp2))


class TestResponseBye(unittest.TestCase):

    def test_bytes(self):
        resp = ResponseBye(b'bye response')
        self.assertEqual(b'* BYE bye response\r\n', bytes(resp))




from __future__ import annotations

import os
import socket
import ssl
from abc import abstractmethod, ABCMeta
from argparse import Namespace
from collections import OrderedDict
from concurrent.futures import ThreadPoolExecutor
from ssl import SSLContext
from typing import Any, TypeVar, Type, Union, Optional, Iterable, Iterator, \
    Sequence, Mapping, Dict
from typing_extensions import Final

from proxyprotocol import ProxyProtocol
from proxyprotocol.version import ProxyProtocolVersion
from pysasl import SASLAuth, AuthenticationCredentials
from pysasl.hashing import HashInterface, get_hash

from .concurrent import Subsystem
from .context import subsystem
from .parsing import Params
from .parsing.commands import Commands

__all__ = ['ConfigT', 'ConfigT_contra', 'BackendCapability', 'IMAPConfig']

#: Type variable with an upper bound of :class:`IMAPConfig`.
ConfigT = TypeVar('ConfigT', bound='IMAPConfig')

#: Contravariant type variable with an upper bound of :class:`IMAPConfig`.
ConfigT_contra = TypeVar('ConfigT_contra', bound='IMAPConfig',
                         contravariant=True)


class BackendCapability(Iterable[bytes]):
    """Declares the IMAP capabilities that the backend supports.

    Args:
        idle: The ``IDLE`` extension is supported.
        object_id: The ``OBJECTID`` extension is supported.
        multi_append: The ``MULTIAPPEND`` extension is supported.
        custom: Optional list of custom capability strings to declare.

    """

    __slots__ = ['_capability']

    def __init__(self, *,
                 idle: bool,
                 object_id: bool,
                 multi_append: bool,
                 custom: Sequence[bytes] = None) -> None:
        super().__init__()
        capability: Dict[bytes, bool] = OrderedDict()
        if idle:
            capability[b'IDLE'] = True
        if object_id:
            capability[b'OBJECTID'] = True
        if multi_append:
            capability[b'MULTIAPPEND'] = True
        if custom is not None:
            for cap in custom:
                capability[cap] = True
        self._capability: Final = tuple(capability.keys())

    def __iter__(self) -> Iterator[bytes]:
        return iter(self._capability)

    def __eq__(self, other) -> bool:
        if isinstance(other, BackendCapability):
            return self._capability == other._capability
        if isinstance(other, tuple):
            return self._capability == other
        if isinstance(other, list):
            return list(self._capability) == other
        return super().__eq__(other)

    def __hash__(self) -> int:
        return hash(self._capability)

    def __repr__(self) -> str:
        return repr([cap.decode('ascii') for cap in self._capability])


class IMAPConfig(metaclass=ABCMeta):
    """Configurable settings that control how the IMAP server operates and what
    extensions it supports.

    Args:
        args: The command-line arguments.
        host: The IMAP server host.
        port: The IMAP server port or service name.
        debug: If true, prints all socket activity to stdout.
        subsystem: The concurrency subsystem in use by the backend.
        ssl_context: SSL context that will be used for opportunistic TLS.
            Alternatively, you can pass extra arguments ``cert_file`` and
            ``key_file`` and an SSL context will be created.
        starttls_enabled: True if opportunistic TLS should be supported.
        hash_context: The hash to use for passwords.
        cpu_subsystem: The subsystem to use for CPU-heavy operations,
            defaulting to a small thread pool.
        preauth_credentials: If given, clients will pre-authenticate on
            connection using these credentials.
        proxy_protocol: The PROXY protocol implementation to use.
        max_append_len: The maximum allowed length of the message body to an
            ``APPEND`` command.
        bad_command_limit: The number of consecutive commands received from
            the client with parsing errors before the client is disconnected.
        disable_idle: Disable the ``IDLE`` capability.
        extra: Additional keywords used for special circumstances.

    Attributes:
        args: The command-line arguments.

    """

    def __init__(self, args: Namespace, *,
                 host: Optional[str],
                 port: Union[str, int],
                 debug: bool = False,
                 subsystem: Subsystem = None,
                 ssl_context: SSLContext = None,
                 tls_enabled: bool = True,
                 secure_auth: bool = True,
                 preauth_credentials: AuthenticationCredentials = None,
                 proxy_protocol: ProxyProtocol = None,
                 hash_context: HashInterface = None,
                 cpu_subsystem: Subsystem = None,
                 max_append_len: Optional[int] = 1000000000,
                 bad_command_limit: Optional[int] = 5,
                 disable_search_keys: Iterable[bytes] = None,
                 disable_idle: bool = False,
                 **extra: Any) -> None:
        super().__init__()
        self.args = args
        self.debug: Final = debug
        self.host: Final = host
        self.port: Final = port
        self.subsystem: Final = subsystem
        self.bad_command_limit: Final = bad_command_limit
        self.disable_search_keys: Final = disable_search_keys or []
        self.hash_context: Final = hash_context or \
            get_hash(passlib_config=args.passlib_cfg)
        self.cpu_subsystem: Final = cpu_subsystem or \
            self._get_cpu_subsystem()
        self._ssl_context = ssl_context or self._load_certs(extra)
        self._tls_enabled = tls_enabled
        self._preauth_credentials = preauth_credentials
        self._proxy_protocol = proxy_protocol or \
            ProxyProtocolVersion.get(args.proxy_protocol)
        self._max_append_len = max_append_len

    @classmethod
    def parse_args(cls, args: Namespace) -> Mapping[str, Any]:
        """Given command-line arguments, return a dictionary of keywords that
        should be passed in to the :class:`IMAPConfig` (or sub-class)
        constructor. Sub-classes should override this method as needed.

        Args:
            args: The arguments parsed from the command-line.

        """
        return {}

    @classmethod
    def from_args(cls: Type[ConfigT], args: Namespace) -> ConfigT:
        """Build and return a new :class:`IMAPConfig` using command-line
        arguments.

        Args:
            args: The arguments parsed from the command-line.

        """
        parsed_args = cls.parse_args(args)
        return cls(args, host=args.host, port=args.port, debug=args.debug,
                   cert_file=args.cert, key_file=args.key,
                   tls_enabled=args.tls, **parsed_args)

    def apply_context(self) -> None:
        """Apply the configured settings to any :mod:`~pymap.context`
        variables.

        """
        if self.subsystem is not None:
            subsystem.set(self.subsystem)

    @property
    @abstractmethod
    def backend_capability(self) -> BackendCapability:
        """Allows backends to declare support for IMAP extensions and other
        capabilities.

        """
        ...

    @classmethod
    def _get_cpu_subsystem(cls) -> Subsystem:
        cpu_count = os.cpu_count() or 1
        cpus_minus_one = max(1, cpu_count - 1)
        executor = ThreadPoolExecutor(max_workers=cpus_minus_one)
        return Subsystem.for_threading(executor)

    @classmethod
    def _load_certs(cls, extra: Mapping[str, Any]) -> SSLContext:
        try:
            cert_file: Optional[str] = os.environ['CERT_FILE']
        except KeyError:
            cert_file = extra.get('cert_file')
        try:
            key_file: Optional[str] = os.environ['KEY_FILE']
        except KeyError:
            key_file = extra.get('key_file')
        ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        if cert_file is not None:
            ssl_context.load_cert_chain(cert_file, key_file)
        return ssl_context

    @property
    def ssl_context(self) -> SSLContext:
        return self._ssl_context

    @property
    def commands(self) -> Commands:
        return Commands()

    @property
    def initial_auth(self) -> SASLAuth:
        if self._tls_enabled:
            return SASLAuth([])
        else:
            return self.tls_auth

    @property
    def tls_auth(self) -> SASLAuth:
        return SASLAuth.defaults()

    @property
    def preauth_credentials(self) -> Optional[AuthenticationCredentials]:
        return self._preauth_credentials

    @property
    def proxy_protocol(self) -> ProxyProtocol:
        return self._proxy_protocol

    @property
    def parsing_params(self) -> Params:
        return Params(max_append_len=self._max_append_len)

    @property
    def greeting(self) -> bytes:
        try:
            fqdn = os.environ['FQDN']
        except KeyError:
            fqdn = socket.getfqdn()
        return b'Server ready ' + fqdn.encode('ascii')

    @property
    def login_capability(self) -> Sequence[bytes]:
        ret = [b'BINARY', b'UIDPLUS', b'MOVE', b'CHILDREN']
        if self._max_append_len is not None:
            ret.append(b'APPENDLIMIT=%i' % self._max_append_len)
        ret.extend(self.backend_capability)
        return ret

    @property
    def initial_capability(self) -> Sequence[bytes]:
        ret = [b'LITERAL+']
        if self._tls_enabled:
            ret.append(b'STARTTLS')
        return ret

    @property
    def max_filter_len(self) -> Optional[int]:
        return self._max_append_len




from __future__ import annotations

import asyncio
import json
import uuid
from argparse import ArgumentParser, Namespace
from contextlib import closing, asynccontextmanager
from datetime import datetime
from functools import partial
from secrets import token_bytes
from typing import Any, Optional, Tuple, Mapping, Sequence, Awaitable, \
    AsyncIterator
from typing_extensions import Final

from aioredis import create_redis, Redis  # type: ignore
from pysasl.creds import AuthenticationCredentials

from pymap.bytes import BytesFormat
from pymap.config import BackendCapability, IMAPConfig
from pymap.context import connection_exit
from pymap.exceptions import AuthorizationFailure, IncompatibleData, \
    NotAllowedError, UserNotFound
from pymap.interfaces.backend import BackendInterface, ServiceInterface
from pymap.interfaces.login import LoginTokenData, LoginInterface, \
    IdentityInterface
from pymap.user import UserMetadata

from .cleanup import CleanupTask
from .filter import FilterSet
from .keys import DATA_VERSION, RedisKey, GlobalKeys, CleanupKeys, \
    NamespaceKeys
from .mailbox import Message, MailboxSet
from ..session import BaseSession

__all__ = ['RedisBackend', 'Config', 'Session']


class RedisBackend(BackendInterface):
    """Defines a backend that uses redis data structures for mailbox storage.

    """

    def __init__(self, login: Login, config: Config) -> None:
        super().__init__()
        self._login = login
        self._config = config

    @property
    def login(self) -> Login:
        return self._login

    @property
    def config(self) -> Config:
        return self._config

    @classmethod
    def add_subparser(cls, name: str, subparsers: Any) -> ArgumentParser:
        parser = subparsers.add_parser(name, help='redis backend')
        parser.add_argument('--address', metavar='URL',
                            default='redis://localhost',
                            help='the redis server address')
        parser.add_argument('--select', metavar='DB', type=int,
                            help='the redis database for mail data')
        parser.add_argument('--separator', metavar='CHAR', default='/',
                            help='the redis key segment separator')
        parser.add_argument('--prefix', metavar='VAL', default='/mail',
                            help='the mail data key prefix')
        parser.add_argument('--users-prefix', metavar='VAL', default='/users',
                            help='the user lookup key prefix')
        parser.add_argument('--users-json', action='store_true',
                            help='the user lookup value contains JSON')
        return parser

    @classmethod
    async def init(cls, args: Namespace) -> Tuple[RedisBackend, Config]:
        config = Config.from_args(args)
        login = Login(config)
        return cls(login, config), config

    async def start(self, services: Sequence[ServiceInterface]) -> Awaitable:
        config = self._config
        global_keys = config._global_keys
        connect_redis = partial(create_redis, config.address)
        cleanup_task = CleanupTask(connect_redis, global_keys)
        tasks = [await cleanup_task.start()]
        for service in services:
            tasks.append(await service.start())
        return asyncio.gather(*tasks)


class Config(IMAPConfig):
    """The config implementation for the redis backend.

    Args:
        args: The command-line arguments.
        address: The redis server address.
        select: The redis database for mail data.
        separator: The redis key segment separator.
        prefix: The prefix for mail data keys.
        users_prefix: The user lookup key prefix.
        users_json: True if the user lookup value contains JSON.

    """

    def __init__(self, args: Namespace, *, address: str, select: Optional[int],
                 separator: bytes, prefix: bytes, users_prefix: bytes,
                 users_json: bool, **extra: Any) -> None:
        super().__init__(args, **extra)
        self._address = address
        self._select = select
        self._separator = separator
        self._prefix = prefix
        self._users_prefix = users_prefix
        self._users_json = users_json

    @property
    def backend_capability(self) -> BackendCapability:
        return BackendCapability(idle=True, object_id=True, multi_append=True)

    @property
    def address(self) -> str:
        """The redis server address. Defaults to a connection to localhost.

        See Also:
            :func:`aioredis.create_connection`

        """
        return self._address

    @property
    def select(self) -> Optional[int]:
        """The redis database for mail data. If given, the `SELECT`_ command is
        called after successful user lookup.

        .. _SELECT: https://redis.io/commands/select

        """
        return self._select

    @property
    def separator(self) -> bytes:
        """The bytestring used to separate segments of composite redis keys."""
        return self._separator

    @property
    def prefix(self) -> bytes:
        """The prefix for mail data keys. This prefix does not apply to
        :attr:`.users_key`.

        """
        return self._prefix

    @property
    def users_prefix(self) -> bytes:
        """The prefix for user lookup keys."""
        return self._users_prefix

    @property
    def users_json(self) -> bool:
        """True if the value from the user lookup key contains a JSON object
        with a ``"password"`` attribute, instead of a redis hash with a
        ``password`` key.

        See Also:
            `redis hashes
            <https://redis.io/topics/data-types-intro#redis-hashes>`_

        """
        return self._users_json

    @property
    def _joiner(self) -> BytesFormat:
        return BytesFormat(self.separator)

    @property
    def _users_root(self) -> RedisKey:
        return RedisKey(self._joiner, [self.users_prefix], {})

    @property
    def _global_keys(self) -> GlobalKeys:
        key = RedisKey(self._joiner, [self.prefix], {})
        return GlobalKeys(key)

    @classmethod
    def parse_args(cls, args: Namespace) -> Mapping[str, Any]:
        return {**super().parse_args(args),
                'address': args.address,
                'select': args.select,
                'separator': args.separator.encode('utf-8'),
                'prefix': args.prefix.encode('utf-8'),
                'users_prefix': args.users_prefix.encode('utf-8'),
                'users_json': args.users_json}


class Session(BaseSession[Message]):
    """The session implementation for the redis backend."""

    resource = __name__

    def __init__(self, redis: Redis, owner: str, config: Config,
                 mailbox_set: MailboxSet, filter_set: FilterSet) -> None:
        super().__init__(owner)
        self._redis = redis
        self._config = config
        self._mailbox_set = mailbox_set
        self._filter_set = filter_set

    @property
    def config(self) -> IMAPConfig:
        return self._config

    @property
    def mailbox_set(self) -> MailboxSet:
        return self._mailbox_set

    @property
    def filter_set(self) -> FilterSet:
        return self._filter_set


class Login(LoginInterface):
    """The login implementation for the redis backend."""

    def __init__(self, config: Config) -> None:
        super().__init__()
        self.config = config

    @classmethod
    async def _connect_redis(cls, address: str) -> Redis:
        redis = await create_redis(address)
        stack = connection_exit.get()
        stack.enter_context(closing(redis))
        return redis

    async def authenticate(self, credentials: AuthenticationCredentials) \
            -> Identity:
        config = self.config
        redis = await self._connect_redis(config.address)
        authcid = credentials.authcid
        token_key: Optional[bytes] = None
        role: Optional[str] = None
        if credentials.authcid_type == 'admin-token':
            authcid = credentials.identity
            role = 'admin'
        try:
            metadata = await Identity(config, redis, authcid, None).get()
        except UserNotFound:
            metadata = UserMetadata(config)
        if 'key' in metadata.params:
            token_key = bytes.fromhex(metadata.params['key'])
        role = role or metadata.role
        await metadata.check_password(credentials, token_key=token_key)
        if role != 'admin' and authcid != credentials.identity:
            raise AuthorizationFailure()
        return Identity(config, redis, credentials.identity, role)


class Identity(IdentityInterface):
    """The identity implementation for the redis backend."""

    def __init__(self, config: Config, redis: Redis, name: str,
                 role: Optional[str]) -> None:
        super().__init__()
        self.config: Final = config
        self._redis: Optional[Redis] = redis
        self._name = name
        self._role = role

    @property
    def name(self) -> str:
        return self._name

    @property
    def redis(self) -> Redis:
        redis = self._redis
        if redis is None:
            # Other methods may not be called after new_session(), since it
            # may have called SELECT on the connection.
            raise RuntimeError()
        return redis

    async def new_token(self, *, expiration: datetime = None) \
            -> Optional[LoginTokenData]:
        metadata = await self.get()
        if 'key' not in metadata.params:
            return None
        key = bytes.fromhex(metadata.params['key'])
        return LoginTokenData(self.name, self.name, key)

    @asynccontextmanager
    async def new_session(self) -> AsyncIterator[Session]:
        config = self.config
        redis = self.redis
        self._redis = None
        if config.select is not None:
            await redis.select(config.select)
        global_keys = config._global_keys
        namespace = await self._get_namespace(redis, global_keys, self.name)
        ns_keys = NamespaceKeys(global_keys, namespace)
        cl_keys = CleanupKeys(global_keys)
        mailbox_set = MailboxSet(redis, ns_keys, cl_keys)
        filter_set = FilterSet(redis, ns_keys)
        try:
            await mailbox_set.add_mailbox('INBOX')
        except ValueError:
            pass
        yield Session(redis, self.name, config, mailbox_set, filter_set)

    async def _get_namespace(self, redis: Redis, global_keys: GlobalKeys,
                             user: str) -> bytes:
        user_key = user.encode('utf-8')
        new_namespace = uuid.uuid4().hex.encode('ascii')
        ns_val = b'%d/%b' % (DATA_VERSION, new_namespace)
        multi = redis.multi_exec()
        multi.hsetnx(global_keys.namespaces, user_key, ns_val)
        multi.hget(global_keys.namespaces, user_key)
        _, ns_val = await multi.execute()
        version, namespace = ns_val.split(b'/', 1)
        if int(version) != DATA_VERSION:
            raise IncompatibleData()
        return namespace

    async def get(self) -> UserMetadata:
        redis = self.redis
        user_bytes = self.name.encode('utf-8')
        user_key = self.config._users_root.end(user_bytes)
        if self.config.users_json:
            json_data = await redis.get(user_key)
            if json_data is None:
                raise UserNotFound(self.name)
            data_dict = json.loads(json_data)
        else:
            data_dict = await redis.hgetall(user_key, encoding='utf-8')
            if data_dict is None:
                raise UserNotFound(self.name)
        return UserMetadata(self.config, **data_dict)

    async def set(self, metadata: UserMetadata) -> None:
        config = self.config
        redis = self.redis
        if self._role != 'admin' and metadata.role:
            raise NotAllowedError('Cannot assign role.')
        user_key = config._users_root.end(self.name.encode('utf-8'))
        user_dict = metadata.to_dict(key=token_bytes().hex())
        if self.config.users_json:
            json_data = json.dumps(user_dict)
            await redis.set(user_key, json_data)
        else:
            multi = redis.multi_exec()
            multi.delete(user_key)
            multi.hmset_dict(user_key, user_dict)
            await multi.execute()

    async def delete(self) -> None:
        config = self.config
        user_key = config._users_root.end(self.name.encode('utf-8'))
        if not await self.redis.delete(user_key):
            raise UserNotFound(self.name)




from __future__ import annotations

import re
from typing import Tuple, Sequence, List, Iterable, Optional, ClassVar

from . import CommandSelect, CommandNoArgs
from .. import Params, Space, EndLine
from ..exceptions import NotParseable
from ..primitives import Atom, ListP
from ..specials import AString, Mailbox, SequenceSet, Flag, FetchAttribute, \
    SearchKey, ExtensionOptions
from ...bytes import rev
from ...flags import FlagOp

__all__ = ['CheckCommand', 'CloseCommand', 'ExpungeCommand', 'CopyCommand',
           'MoveCommand', 'FetchCommand', 'StoreCommand', 'SearchCommand',
           'UidCommand', 'UidCopyCommand', 'UidMoveCommand',
           'UidExpungeCommand', 'UidFetchCommand', 'UidSearchCommand',
           'UidStoreCommand', 'IdleCommand']


class CheckCommand(CommandNoArgs, CommandSelect):
    """The ``CHECK`` command initiates an implementation-specific backend
    synchronization for the selected mailbox.

    See Also:
       `RFC 3501 6.4.1. <https://tools.ietf.org/html/rfc3501#section-6.4.1>`_

    """

    command = b'CHECK'


class CloseCommand(CommandNoArgs, CommandSelect):
    """The ``CLOSE`` command closes a selected mailbox.

    See Also:
        `RFC 3501 6.4.2. <https://tools.ietf.org/html/rfc3501#section-6.4.2>`_

    """

    command = b'CLOSE'


class ExpungeCommand(CommandSelect):
    """The ``EXPUNGE`` command permanently erases all messages in the selected
    mailbox that contain the ``\\Deleted`` flag.

    See Also:
        `RFC 3501 6.4.3 <https://tools.ietf.org/html/rfc3501#section-6.4.3>`_
        `RFC 4315 2.1 <https://tools.ietf.org/html/rfc4315#section-2.1>`_

    Args:
        tag: The command tag.
        uid_set: Only the messages in the given UID set should be expunged.

    """

    command = b'EXPUNGE'
    uid: ClassVar[bool] = False

    def __init__(self, tag: bytes, uid_set: SequenceSet = None) -> None:
        super().__init__(tag)
        self.uid_set = uid_set

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[ExpungeCommand, memoryview]:
        uid_set: Optional[SequenceSet] = None
        if params.uid:
            _, buf = Space.parse(buf, params)
            uid_set, buf = SequenceSet.parse(buf, params)
        _, buf = EndLine.parse(buf, params)
        return cls(params.tag, uid_set), buf


class CopyCommand(CommandSelect):
    """The ``COPY`` command copies messages from the selected mailbox to the
    end of the destination mailbox.

    See Also:
        `RFC 3501 6.4.7. <https://tools.ietf.org/html/rfc3501#section-6.4.7>`_

    Args:
        tag: The command tag.
        seq_set: The sequence set of the messages to copy.
        mailbox: The destination mailbox.

    """

    command = b'COPY'
    uid: ClassVar[bool] = False

    def __init__(self, tag: bytes, seq_set: SequenceSet,
                 mailbox: Mailbox) -> None:
        super().__init__(tag)
        self.sequence_set = seq_set
        self.mailbox_obj = mailbox

    @property
    def mailbox(self) -> str:
        return str(self.mailbox_obj)

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[CopyCommand, memoryview]:
        _, buf = Space.parse(buf, params)
        seq_set, buf = SequenceSet.parse(buf, params)
        _, buf = Space.parse(buf, params)
        mailbox, buf = Mailbox.parse(buf, params)
        _, buf = EndLine.parse(buf, params)
        return cls(params.tag, seq_set, mailbox), buf


class MoveCommand(CommandSelect):
    """The ``MOVE`` command moves messages from the selected mailbox to the
    end of the destination mailbox.

    See Also:
        `RFC 6851 <https://tools.ietf.org/html/rfc6851>`_

    Args:
        tag: The command tag.
        seq_set: The sequence set of the messages to move.
        mailbox: The destination mailbox.

    """

    command = b'MOVE'
    uid: ClassVar[bool] = False

    def __init__(self, tag: bytes, seq_set: SequenceSet,
                 mailbox: Mailbox) -> None:
        super().__init__(tag)
        self.sequence_set = seq_set
        self.mailbox_obj = mailbox

    @property
    def mailbox(self) -> str:
        return str(self.mailbox_obj)

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[MoveCommand, memoryview]:
        _, buf = Space.parse(buf, params)
        seq_set, buf = SequenceSet.parse(buf, params)
        _, buf = Space.parse(buf, params)
        mailbox, buf = Mailbox.parse(buf, params)
        _, buf = EndLine.parse(buf, params)
        return cls(params.tag, seq_set, mailbox), buf


class FetchCommand(CommandSelect):
    """The ``FETCH`` command fetches message data from the selected mailbox.
    What data is fetched can be controlled in depth by a set of fetch
    attributes given in the command.

    See Also:
        `RFC 3501 6.4.5. <https://tools.ietf.org/html/rfc3501#section-6.4.5>`_

    Args:
        tag: The command tag.
        seq_set: The sequence set of the messages to fetch.
        attr_list: The message attributes to fetch.

    """

    command = b'FETCH'
    uid: ClassVar[bool] = False

    def __init__(self, tag: bytes, seq_set: SequenceSet,
                 attr_list: Sequence[FetchAttribute],
                 options: ExtensionOptions = None) -> None:
        super().__init__(tag)
        self.sequence_set = seq_set
        self.attributes = attr_list
        self.options = options or ExtensionOptions.empty()

    @classmethod
    def _check_macros(cls, buf: memoryview, params: Params) \
            -> Tuple[Sequence[FetchAttribute], memoryview]:
        atom, after = Atom.parse(buf, params)
        macro = atom.value.upper()
        if macro == b'ALL':
            attrs = [FetchAttribute(b'FLAGS'),
                     FetchAttribute(b'INTERNALDATE'),
                     FetchAttribute(b'RFC822.SIZE'),
                     FetchAttribute(b'ENVELOPE')]
            return attrs, after
        elif macro == b'FULL':
            attrs = [FetchAttribute(b'FLAGS'),
                     FetchAttribute(b'INTERNALDATE'),
                     FetchAttribute(b'RFC822.SIZE'),
                     FetchAttribute(b'ENVELOPE'),
                     FetchAttribute(b'BODY')]
            return attrs, after
        elif macro == b'FAST':
            attrs = [FetchAttribute(b'FLAGS'),
                     FetchAttribute(b'INTERNALDATE'),
                     FetchAttribute(b'RFC822.SIZE')]
            return attrs, after
        raise NotParseable(buf)

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[FetchCommand, memoryview]:
        _, buf = Space.parse(buf, params)
        seq_set, buf = SequenceSet.parse(buf, params)
        _, buf = Space.parse(buf, params)
        attr_list: Sequence[FetchAttribute] = []
        try:
            attr_list, buf = cls._check_macros(buf, params)
        except NotParseable:
            pass
        try:
            attr, buf = FetchAttribute.parse(buf, params)
            attr_list = [attr]
        except NotParseable:
            pass
        if not attr_list:
            params_copy = params.copy(list_expected=[FetchAttribute])
            attr_list_p, buf = ListP.parse(buf, params_copy)
            attr_list = attr_list_p.get_as(FetchAttribute)
        if params.uid:
            attr_list = list(attr_list) + [FetchAttribute(b'UID')]
        options, buf = ExtensionOptions.parse(buf, params)
        _, buf = EndLine.parse(buf, params)
        return cls(params.tag, seq_set, attr_list, options), buf


class StoreCommand(CommandSelect):
    """The ``STORE`` command updates the flags of a set of messages in the
    selected mailbox.

    See Also:
        `RFC 3501 6.4.6. <https://tools.ietf.org/html/rfc3501#section-6.4.6>`_

    Args:
        tag: The command tag.
        seq_set: The sequence set of the messages to fetch.
        flags: The flag set operand.
        mode: The type of update operation.

    """

    command = b'STORE'
    uid: ClassVar[bool] = False

    _info_pattern = rev.compile(br'^([+-]?)FLAGS(\.SILENT)?$', re.I)
    _modes = {b'': FlagOp.REPLACE, b'+': FlagOp.ADD, b'-': FlagOp.DELETE}

    def __init__(self, tag: bytes, seq_set: SequenceSet,
                 flags: Iterable[Flag], mode: FlagOp,
                 silent: bool, options: ExtensionOptions = None) -> None:
        super().__init__(tag)
        self.sequence_set = seq_set
        self.flag_set = frozenset(flags)
        self.mode = mode
        self.silent = silent
        self.options = options or ExtensionOptions.empty()

    @classmethod
    def _parse_store_info(cls, buf: memoryview, params: Params) \
            -> Tuple[FlagOp, bool, memoryview]:
        info, after = Atom.parse(buf, params)
        match = cls._info_pattern.match(info.value)
        if not match:
            raise NotParseable(buf)
        mode = cls._modes[match.group(1)]
        silent = bool(match.group(2))
        return mode, silent, after

    @classmethod
    def _parse_flag_list(cls, buf: memoryview, params: Params) \
            -> Tuple[Sequence[Flag], memoryview]:
        try:
            params_copy = params.copy(list_expected=[Flag])
            flag_list_p, buf = ListP.parse(buf, params_copy)
        except NotParseable:
            pass
        else:
            return flag_list_p.get_as(Flag), buf
        flag_list: List[Flag] = []
        while True:
            try:
                flag, buf = Flag.parse(buf, params)
                flag_list.append(flag)
                _, buf = Space.parse(buf, params)
            except NotParseable:
                return flag_list, buf

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[StoreCommand, memoryview]:
        _, buf = Space.parse(buf, params)
        seq_set, buf = SequenceSet.parse(buf, params)
        options, buf = ExtensionOptions.parse(buf, params)
        _, buf = Space.parse(buf, params)
        mode, silent, buf = cls._parse_store_info(buf, params)
        _, buf = Space.parse(buf, params)
        flag_list, buf = cls._parse_flag_list(buf, params)
        _, buf = EndLine.parse(buf, params)
        return cls(params.tag, seq_set, flag_list, mode, silent, options), buf


class SearchCommand(CommandSelect):
    """The ``SEARCH`` command searches the messages in the selected mailbox
    based on a set of search criteria.

    See Also:
        `RFC 3501 6.4.4. <https://tools.ietf.org/html/rfc3501#section-6.4.4>`_

    Args:
        tag: The command tag.
        keys: The search keys.
        charset: The charset in use by the search keys.

    """

    command = b'SEARCH'
    uid: ClassVar[bool] = False

    def __init__(self, tag: bytes, keys: Iterable[SearchKey],
                 charset: Optional[str],
                 options: ExtensionOptions = None) -> None:
        super().__init__(tag)
        self.keys = frozenset(keys)
        self.charset = charset
        self.options = options or ExtensionOptions.empty()

    @classmethod
    def _parse_charset(cls, buf: memoryview, params: Params) \
            -> Tuple[Optional[str], memoryview]:
        try:
            _, after = Space.parse(buf, params)
            atom, after = Atom.parse(after, params)
        except NotParseable:
            pass
        else:
            if atom.value.upper() == b'CHARSET':
                _, after = Space.parse(after, params)
                string, after = AString.parse(after, params)
                charset = str(string.value, 'ascii')
                try:
                    b' '.decode(charset)
                except LookupError:
                    raise NotParseable(buf, b'BADCHARSET')
                return charset, after
        return None, buf

    @classmethod
    def _parse_options(cls, buf: memoryview, params: Params) \
            -> Tuple[ExtensionOptions, memoryview]:
        start = cls._whitespace_length(buf)
        if buf[start:start + 6] == b'RETURN':
            return ExtensionOptions.parse(buf[start + 6:], params)
        else:
            options, _ = ExtensionOptions.parse(memoryview(b''), params)
            return options, buf

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[SearchCommand, memoryview]:
        options, buf = cls._parse_options(buf, params)
        charset, buf = cls._parse_charset(buf, params)
        search_keys = []
        while True:
            try:
                _, buf = Space.parse(buf, params)
                key, buf = SearchKey.parse(buf, params.copy(charset=charset))
                search_keys.append(key)
            except NotParseable:
                if not search_keys:
                    raise
                break
        _, buf = EndLine.parse(buf, params)
        return cls(params.tag, search_keys, charset, options), buf


class UidCommand(CommandSelect):
    """The ``UID`` command precedes one of the ``COPY``, ``EXPUNGE``,
    ``FETCH``, ``SEARCH``, or ``STORE`` commands and indicates that the
    command interacts with message UIDs instead of sequence numbers. Refer
    to the RFC section for a complete description.

    See Also:
        `RFC 3501 6.4.8 <https://tools.ietf.org/html/rfc3501#section-6.4.8>`_
        `RFC 4315 2.1 <https://tools.ietf.org/html/rfc4315#section-2.1>`_

    """

    command = b'UID'
    compound = True

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[UidCommand, memoryview]:
        raise TypeError(cls)  # never parsed directly


class UidCopyCommand(CopyCommand):
    """The ``UID COPY`` variant of the ``COPY`` command, which uses message
    UIDs instead of sequence numbers.

    """

    command = b'UID COPY'
    delegate = CopyCommand
    uid = True

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[UidCopyCommand, memoryview]:
        ret, buf = super().parse(buf, params.copy(uid=True))
        if not isinstance(ret, UidCopyCommand):
            raise TypeError(ret)
        return ret, buf


class UidMoveCommand(MoveCommand):
    """The ``UID MOVE`` variant of the ``MOVE`` command, which uses message
    UIDs instead of sequence numbers.

    """

    command = b'UID MOVE'
    delegate = MoveCommand
    uid = True

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[UidMoveCommand, memoryview]:
        ret, buf = super().parse(buf, params.copy(uid=True))
        if not isinstance(ret, UidMoveCommand):
            raise TypeError(ret)
        return ret, buf


class UidExpungeCommand(ExpungeCommand):
    """The ``UID EXPUNGE`` variant of the ``EXPUNGE`` command, which uses
    message UIDs instead of sequence numbers.

    """

    command = b'UID EXPUNGE'
    delegate = ExpungeCommand
    uid = True

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[UidExpungeCommand, memoryview]:
        ret, buf = super().parse(buf, params.copy(uid=True))
        if not isinstance(ret, UidExpungeCommand):
            raise TypeError(ret)
        return ret, buf


class UidFetchCommand(FetchCommand):
    """The ``UID FETCH`` variant of the ``FETCH`` command, which uses message
    UIDs instead of sequence numbers.

    """

    command = b'UID FETCH'
    delegate = FetchCommand
    uid = True

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[UidFetchCommand, memoryview]:
        ret, buf = super().parse(buf, params.copy(uid=True))
        if not isinstance(ret, UidFetchCommand):
            raise TypeError(ret)
        return ret, buf


class UidSearchCommand(SearchCommand):
    """The ``UID SEARCH`` variant of the ``SEARCH`` command, which uses message
    UIDs instead of sequence numbers.

    """

    command = b'UID SEARCH'
    delegate = SearchCommand
    uid = True

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[UidSearchCommand, memoryview]:
        ret, buf = super().parse(buf, params.copy(uid=True))
        if not isinstance(ret, UidSearchCommand):
            raise TypeError(ret)
        return ret, buf


class UidStoreCommand(StoreCommand):
    """The ``UID STORE`` variant of the ``STORE`` command, which uses message
    UIDs instead of sequence numbers.

    """

    command = b'UID STORE'
    delegate = StoreCommand
    uid = True

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[UidStoreCommand, memoryview]:
        ret, buf = super().parse(buf, params.copy(uid=True))
        if not isinstance(ret, UidStoreCommand):
            raise TypeError(ret)
        return ret, buf


class IdleCommand(CommandSelect):
    """The ``IDLE`` command waits for the continuation string ``DONE`` from the
    client. During this wait, the server may be sending untagged responses
    indicating concurrent updates to the mailbox.

    Parsing this command is a special case. The continuation string ``DONE`` is
    actually parsed by the :meth:`.parse_done` method. The :meth:`.parse`
    only parses the ``IDLE <CRLF>`` portion, and does not raise a
    :exc:`~pymap.exceptions.RequiresContinuation` exception.

    See Also:
        `RFC 2177 <https://tools.ietf.org/html/rfc2177>`_

    """

    command = b'IDLE'

    #: The string used to end the command.
    continuation = b'DONE'

    _pattern = rev.compile(br'^(.*?)\r?\n')

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[IdleCommand, memoryview]:
        _, buf = EndLine.parse(buf, params)
        return cls(params.tag), buf

    def parse_done(self, buf: memoryview) -> Tuple[bool, memoryview]:
        """Parse the continuation line sent by the client to end the ``IDLE``
        command.

        Args:
            buf: The continuation line to parse.

        """
        match = self._pattern.match(buf)
        if not match:
            raise NotParseable(buf)
        done = match.group(1).upper() == self.continuation
        buf = buf[match.end(0):]
        return done, buf




from __future__ import annotations

import hashlib
import os.path
from contextlib import closing
from typing import Generic, TypeVar, Any, Union, Optional, Tuple, Sequence

import msgpack  # type: ignore
from aioredis import Redis, ReplyError  # type: ignore
from pkg_resources import resource_stream

__all__ = ['ScriptBase']

_Val = Union[int, bytes]
_RetT = TypeVar('_RetT')


class ScriptBase(Generic[_RetT]):
    """The base class for redis scripts.

    Args:
        name: The script name, which should correspond to a ``.lua`` file.

    """

    __slots__ = ['_name', '_sha']

    def __init__(self, name: str) -> None:
        super().__init__()
        self._name = name
        self._sha, _ = self._load()

    def _load(self) -> Tuple[str, bytes]:
        fname = os.path.join('lua', f'{self._name}.lua')
        with closing(resource_stream(__name__, fname)) as script:
            data = script.read()
        return hashlib.sha1(data).hexdigest(), data

    def _convert(self, ret: Any) -> _RetT:
        return ret

    def _pack(self, val: Any) -> bytes:
        return msgpack.packb(val)

    def _maybe_int(self, val: bytes) -> Optional[int]:
        if val == b'':
            return None
        else:
            return int(val)

    async def eval(self, redis: Redis, keys: Sequence[bytes],
                   args: Sequence[_Val]) -> _RetT:
        """Run the script.

        Notes:
            This attempts the EVALSHA command first, then falls back to EVAL if
            a NOSCRIPT error is returned.

        Raises:
            :class:`~asyncio.ReplyError`

        """
        try:
            ret = await redis.evalsha(self._sha, keys, args)
        except ReplyError as exc:
            if 'NOSCRIPT' in str(exc):
                self._sha, data = self._load()
                ret = await redis.eval(data, keys, args)
            else:
                raise
        return self._convert(ret)




from __future__ import annotations

import base64
from email.headerregistry import ContentTypeHeader
from email.policy import SMTP
from itertools import chain, islice
from typing import Any, Optional, Tuple, Iterable, Sequence, Mapping, \
    List, Dict
from typing_extensions import Final

from .parsed import ParsedHeaders
from ._util import whitespace, find_any, get_raw
from ..bytes import WriteStream, Writeable

__all__ = ['MessageContent', 'MessageHeader', 'MessageBody']

_default_type: Final = 'text/plain'

_Line = Tuple[int, int, int]
_Lines = Sequence[_Line]
_Folded = Sequence[Tuple[str, _Lines]]


class MessageContent(Writeable):
    """Contains the message content, parsed for IMAP processing.

    Args:
        data: The message literal.
        header: The parsed message header.
        body: The parsed message body.

    Attributes:
        lines: The number of lines in the message content.

    """

    __slots__ = ['_raw', 'lines', 'header', 'body', '__weakref__']

    def __init__(self, data: bytes, header: MessageHeader,
                 body: MessageBody) -> None:
        super().__init__()
        self._raw = get_raw(memoryview(data), header._lines, body._lines)
        self.lines: Final = header.lines + body.lines - 1
        self.header: Final = header
        self.body: Final = body

    def walk(self) -> Iterable[MessageContent]:
        """Iterate through the message and all its nested sub-parts in the
        order they occur.

        """
        if self.body.has_nested:
            return chain([self], *(part.walk() for part in self.body.nested))
        else:
            return [self]

    @property
    def is_rfc822(self) -> bool:
        """True if the content-type of the message is ``message/rfc822``."""
        ct_hdr = self.header.parsed.content_type
        if ct_hdr is None:
            return False
        else:
            return ct_hdr.content_type == 'message/rfc822'

    @property
    def json(self) -> Mapping[str, Any]:
        """A dictionary that can be serialized (e.g. with :mod:`json`), so that
        this object may be re-created without parsing.

        See Also:
            :meth:`.from_json`

        """
        return {'header': self.header.json,
                'body': self.body.json}

    @classmethod
    def from_json(cls, data: bytes, json: Mapping[str, Any]) -> MessageContent:
        """Recover the parsed message content without re-parsing, using the
        original raw data and the :attr:`.json`.

        In this example, ``content1`` and ``content2`` should be equivalent::

            message = b'...'
            content1 = MessageContent.parse(message)
            content2 = MessageContent.from_json(message, content1.json)

        Args:
            data: The bytestring that was parsed.
            json: The :attr:`.json` of a previously parsed message content.

        """
        header = MessageHeader.from_json(data, json['header'])
        body = MessageBody.from_json(data, json['body'])
        return cls(data, header, body)

    @classmethod
    def parse(cls, data: bytes) -> MessageContent:
        """Parse the bytestring into message content.

        Args:
            data: The bytestring to parse.

        """
        lines = cls._find_lines(data)
        view = memoryview(data)
        return cls._parse(data, view, lines)

    @classmethod
    def _parse(cls, data: bytes, view: memoryview, lines: _Lines) \
            -> MessageContent:
        header_lines, body_lines = cls._split_lines(data, lines)
        header = MessageHeader._parse(data, view, header_lines)
        content_type = header.parsed.content_type
        body = MessageBody._parse(data, view, body_lines, content_type)
        return cls(data, header, body)

    @classmethod
    def _find_lines(cls, data: bytes) -> _Lines:
        start = 0
        end = len(data)
        ret: List[_Line] = []
        while True:
            idx = data.find(b'\n', start, end)
            if idx < 0:
                ret.append((start, end, end))
                break
            next_start = idx + 1
            idx_minus_1 = idx - 1
            if idx_minus_1 >= start and data[idx_minus_1] == 0x0d:
                idx = idx_minus_1
            ret.append((start, idx, next_start))
            start = next_start
        return ret

    @classmethod
    def _split_lines(cls, data: bytes, lines: _Lines) -> Tuple[_Lines, _Lines]:
        for i, line in enumerate(lines):
            start, end, _ = line
            ws_end = find_any(data, whitespace, start, end, False, False)
            if ws_end < 0:
                return lines[0:i + 1], lines[i + 1:]
        return [], lines

    def write(self, writer: WriteStream) -> None:
        writer.write(bytes(self))

    def __len__(self) -> int:
        return len(self._raw)

    def __bytes__(self) -> bytes:
        return bytes(self._raw)


class MessageHeader(Writeable):
    """The message header. Contains lines in the form of ``Header: value\\n``,
    possibly folded onto multiple lines where subsequent lines start with
    whitespace.

    Attributes:
        folded: A list of headers, as they occurred in the original data, as
            tuples of the lower-cased header name and the full header value,
            including the header name and any extra folded lines.
        parsed: The message headers, as a dictionary-like object that parses
            headers on-demand.

    """

    __slots__ = ['_raw', '_lines', '_folded', 'folded', 'parsed']

    def __init__(self, data: bytes, lines: _Lines, folded: _Folded) -> None:
        super().__init__()
        view = memoryview(data)
        self._raw = get_raw(view, lines)
        self._lines = lines
        self._folded = folded
        self.folded: Final = self._get_folded(view, folded)
        self.parsed: Final = self._get_parsed(data, folded)

    @property
    def lines(self) -> int:
        """The number of lines in the message header."""
        return len(self._lines)

    @property
    def json(self) -> Mapping[str, Any]:
        """A dictionary that can be serialized (e.g. with :mod:`json`), so that
        this object may be re-created without parsing.

        See Also:
            :meth:`.from_json`

        """
        return {'lines': self._lines,
                'folded': self._folded}

    @classmethod
    def from_json(cls, data: bytes, json: Mapping[str, Any]) -> MessageHeader:
        """Recover the parsed message header without re-parsing, using the
        original raw data and the :attr:`.json`.

        See Also:
            :meth:`MessageContent.from_json`

        Args:
            data: The bytestring that was parsed.
            json: The :attr:`.json` of a previously parsed message content.

        """
        lines: _Lines = json['lines']
        folded: _Folded = json['folded']
        return cls(data, lines, folded)

    @classmethod
    def empty(cls) -> MessageHeader:
        """Return an empty header object."""
        return cls(b'', [], [])

    @classmethod
    def _get_folded(cls, view: memoryview, folded: _Folded) \
            -> Sequence[Tuple[bytes, memoryview]]:
        return [(cls._to_bytes(key), get_raw(view, lines))
                for key, lines in folded]

    @classmethod
    def _get_parsed(cls, data: bytes, folded: _Folded) -> ParsedHeaders:
        header_map: Dict[bytes, List[List[bytes]]] = {}
        for key, lines in folded:
            name = cls._to_bytes(key)
            values = header_map.setdefault(name, [])
            values.append([data[start:end] for start, end, _ in lines])
        return ParsedHeaders(header_map)

    @classmethod
    def _to_bytes(cls, key: str) -> bytes:
        return base64.b64decode(key)

    @classmethod
    def _to_str(cls, key: bytes) -> str:
        return base64.b64encode(key).decode('ascii')

    @classmethod
    def _parse(cls, data: bytes, view: memoryview,
               lines: _Lines) -> MessageHeader:
        folds = cls._find_folds(data, lines)
        folded = cls._find_folded(data, view, folds)
        return cls(data, lines, folded)

    @classmethod
    def _find_folds(cls, data: bytes, lines: _Lines) -> Sequence[_Lines]:
        ret: List[List[Tuple[int, int, int]]] = []
        if not lines:
            return []
        for line in islice(lines, len(lines) - 1):
            start, end, _ = line
            length = end - start
            if length >= 1 and data[start] in whitespace:
                if ret:
                    ret[-1].append(line)
            else:
                ret.append([line])
        return ret

    @classmethod
    def _find_folded(cls, data: bytes, view: memoryview,
                     folds: Sequence[_Lines]) -> _Folded:
        folded: List[Tuple[str, _Lines]] = []
        for group in folds:
            start, end, _ = group[0]
            colon = data.find(b':', start, end)
            if colon < 0:
                continue
            name = data[start:colon].strip().lower()
            folded.append((cls._to_str(name), group))
        return folded

    def write(self, writer: WriteStream) -> None:
        writer.write(bytes(self))

    def __len__(self) -> int:
        return len(self._raw)

    def __bytes__(self) -> bytes:
        return bytes(self._raw)


class MessageBody(Writeable):
    """The message body, starting immediately after the header. The body may
    contain nested sub-parts, which are each valid :class:`MessageContent`
    objects.

    Attributes:
        content_type: The content type of the message body.

    """

    __slots__ = ['_raw', '_lines', '_nested', 'content_type']

    def __init__(self, data: bytes, lines: _Lines,
                 content_type: ContentTypeHeader,
                 nested: Sequence[MessageContent]) -> None:
        super().__init__()
        self._raw = get_raw(memoryview(data), lines)
        self._lines = lines
        self._nested = nested
        self.content_type: Final = content_type

    @property
    def lines(self) -> int:
        """The number of lines in the message body."""
        return len(self._lines)

    @property
    def has_nested(self) -> bool:
        """True if the message body is composed of nested sub-parts."""
        return len(self.nested) > 0

    @property
    def nested(self) -> Sequence[MessageContent]:
        """If :attr:`.has_nested` is True, contains the list of sub-parts."""
        return self._nested

    @property
    def json(self) -> Mapping[str, Any]:
        """A dictionary that can be serialized (e.g. with :mod:`json`), so that
        this object may be re-created without parsing.

        See Also:
            :meth:`.from_json`

        """
        return {'lines': self._lines,
                'content_type': str(self.content_type),
                'nested': [part.json for part in self._nested]}

    @classmethod
    def from_json(cls, data: bytes, json: Mapping[str, Any]) -> MessageBody:
        """Recover the parsed message body without re-parsing, using the
        original raw data and the :attr:`.json`.

        See Also:
            :meth:`MessageContent.from_json`

        Args:
            data: The bytestring that was parsed.
            json: The :attr:`.json` of a previously parsed message content.

        """
        lines: _Lines = json['lines']
        content_type = cls._parse_content_type(json['content_type'])
        nested = [MessageContent.from_json(data, part)
                  for part in json['nested']]
        return cls(data, lines, content_type, nested)

    @classmethod
    def empty(cls) -> MessageBody:
        """Return an empty body object."""
        content_type = cls._parse_content_type(_default_type)
        return cls(b'', [], content_type, [])

    @classmethod
    def _parse(cls, data: bytes, view: memoryview, lines: _Lines,
               content_type: Optional[ContentTypeHeader]) -> MessageBody:
        if content_type is None:
            content_type = cls._parse_content_type(_default_type)
        maintype = content_type.maintype
        if maintype == 'multipart':
            boundary = cls._get_boundary(content_type)
            if boundary:
                return cls._parse_multipart(
                    data, view, lines, content_type, boundary)
        elif maintype == 'message' and content_type.subtype == 'rfc822':
            return cls._parse_rfc822(data, view, lines, content_type)
        return cls(data, lines, content_type, [])

    @classmethod
    def _parse_content_type(cls, header: str) -> ContentTypeHeader:
        return SMTP.header_fetch_parse('Content-Type', header)  # type: ignore

    @classmethod
    def _get_boundary(cls, content_type: ContentTypeHeader) -> Optional[bytes]:
        try:
            boundary = content_type.params['boundary']
        except KeyError:
            pass
        else:
            if boundary:
                try:
                    return boundary.encode('ascii')
                except UnicodeError:
                    pass
        return None

    @classmethod
    def _parse_rfc822(cls, data: bytes, view: memoryview, lines: _Lines,
                      content_type: ContentTypeHeader) -> MessageBody:
        subpart = MessageContent._parse(data, view, lines)
        return cls(data, lines, content_type, [subpart])

    @classmethod
    def _parse_multipart(cls, data: bytes, view: memoryview, lines: _Lines,
                         content_type: ContentTypeHeader,
                         boundary: bytes) -> MessageBody:
        parts = cls._find_parts(data, view, lines, boundary)
        nested: List[MessageContent] = []
        for part_lines in parts:
            sub_content = MessageContent._parse(data, view, part_lines)
            nested.append(sub_content)
        return cls(data, lines, content_type, nested)

    @classmethod
    def _find_parts(cls, data: bytes, view: memoryview, lines: _Lines,
                    boundary: bytes) -> Sequence[_Lines]:
        ret: List[List[_Line]] = []
        part_match = (b'--%s' % boundary, b'--%s' % boundary)
        stop_match = (b'--%s--' % boundary, b'--%s--' % boundary)
        for line in lines:
            start, end, _ = line
            line_view = view[start:end]
            if any(line_view == m for m in stop_match):
                break
            elif any(line_view == m for m in part_match):
                ret.append([])
            elif ret:
                ret[-1].append(line)
        return ret

    def write(self, writer: WriteStream) -> None:
        writer.write(bytes(self))

    def __len__(self) -> int:
        return len(self._raw)

    def __bytes__(self) -> bytes:
        return bytes(self._raw)
